# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
#
# Copyright (c) 2020, NVIDIA CORPORATION.
# Some functions in this file are adapted from following sources:
# - clip_grad_norm_fp32 : https://github.com/NVIDIA/Megatron-LM/blob/5ac5571ba0265af4c491ee0af1508ca7589450c6/megatron/optimizer/clip_grads.py#L29
# - check_overflow_for_grads : https://github.com/NVIDIA/Megatron-LM/blob/5ac5571ba0265af4c491ee0af1508ca7589450c6/megatron/optimizer/optimizer.py#L341
# --------------------------------------------------------------------------

import warnings

import torch
from numpy import inf

from ._multi_tensor_apply import MultiTensorApply

multi_tensor_applier = MultiTensorApply(2048 * 32)


class FP16OptimizerModifier:
    def __init__(self, optimizer) -> None:
        super().__init__()
        self._optimizer = optimizer

    def apply(self):
        if self.can_be_modified():
            self.override_function()

    def check_requirements(self, required_funcs, require_apex=False, require_torch_non_finite_check=False):
        try:
            if require_apex is True:
                import amp_C  # noqa: F401
                from apex import amp  # noqa: F401
            if require_torch_non_finite_check is True:
                _ = torch._amp_foreach_non_finite_check_and_unscale_
        except Exception:
            warnings.warn("Skip modifying optimizer because of Apex or torch_non_finite_check not found.", UserWarning)
            return False

        if required_funcs:
            for func_name in required_funcs:
                func = getattr(self._optimizer, func_name, None)
                if not func or not callable(func):
                    warnings.warn(
                        "Skip modifying optimizer because of specific function not found in optimizer.", UserWarning
                    )
                    return False
        return True


def check_overflow(params):
    grad_data = [p.grad.data for p in params if p.grad is not None]
    return check_overflow_for_grads(grad_data)


def check_overflow_for_grads(grad_data):
    found_inf = torch.cuda.FloatTensor([0.0])
    scaler = torch.cuda.FloatTensor([1.0])
    # Unscale and set found inf/nan
    torch._amp_foreach_non_finite_check_and_unscale_(grad_data, found_inf, scaler)

    # Check for nan.
    overflow = found_inf.item() > 0
    return overflow


def clip_grad_norm_fp32(
    parameters, max_norm, norm_type, get_horizontal_model_parallel_rank=None, get_horizontal_model_parallel_group=None
):
    from onnxruntime.training.ortmodule.torch_cpp_extensions import fused_ops

    horizontal_model_parallel_grad_norm_aggregation = False
    if get_horizontal_model_parallel_rank and get_horizontal_model_parallel_group:
        horizontal_model_parallel_grad_norm_aggregation = True

    def param_is_not_tensor_parallel_duplicate(param):
        is_mp_tensor = hasattr(param, "model_parallel") and param.model_parallel
        return is_mp_tensor or (get_horizontal_model_parallel_rank() == 0)

    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]

    # Filter parameters based on:
    #   - grad should not be none
    #   - should not be a replica due to tensor model parallelism
    grads_for_norm = []
    for param in parameters:
        grad_not_none = param.grad is not None
        grad = param.grad.detach()
        if grad_not_none:
            # Make sure the grads are in fp32
            assert param.grad.type() == "torch.cuda.FloatTensor"
            if horizontal_model_parallel_grad_norm_aggregation:
                is_not_tp_duplicate = param_is_not_tensor_parallel_duplicate(param)
                if grad_not_none and is_not_tp_duplicate:
                    grads_for_norm.append(grad)
            else:
                grads_for_norm.append(grad)

    # Norm parameters.
    max_norm = float(max_norm)
    norm_type = float(norm_type)
    total_norm = 0.0

    dummy_overflow_buf = torch.cuda.IntTensor([0])
    # Calculate norm.
    if norm_type == inf:
        total_norm = max(grad.abs().max() for grad in grads_for_norm)
        if horizontal_model_parallel_grad_norm_aggregation:
            total_norm_cuda = torch.cuda.FloatTensor([float(total_norm)])

            # Take max across all model-parallel GPUs.
            torch.distributed.all_reduce(
                total_norm_cuda, op=torch.distributed.ReduceOp.MAX, group=get_horizontal_model_parallel_group()
            )
            total_norm = total_norm_cuda[0].item()

    else:
        if norm_type == 2.0:
            # Use apex's multi-tensor applier for efficiency reasons.
            # Multi-tensor applier takes a function and a list of list
            # and performs the operation on that list all in one kernel.
            grad_norm, _ = multi_tensor_applier(
                fused_ops.multi_tensor_l2norm,
                dummy_overflow_buf,
                [fused_ops.TorchTensorVector(grads_for_norm)],
                False,  # no per-parameter norm
            )

            if not horizontal_model_parallel_grad_norm_aggregation:
                return grad_norm.item()

            # Since we will be summing across data parallel groups,
            # we need the pow(norm-type).
            total_norm = grad_norm**norm_type

        else:
            for grad in grads_for_norm:
                # torch.norm is deprecated and moved to torch.linalg.norm
                # with a different signature
                # see https://pytorch.org/docs/stable/generated/torch.norm.html
                if norm_type in {"fro", "nuc"}:
                    grad_norm = torch.linalg.matrix_norm(grad, norm_type)
                else:
                    grad_norm = torch.linalg.norm(grad, norm_type)
                total_norm += grad_norm**norm_type

        if horizontal_model_parallel_grad_norm_aggregation:
            # Sum across all model-parallel GPUs.
            torch.distributed.all_reduce(
                total_norm, op=torch.distributed.ReduceOp.SUM, group=get_horizontal_model_parallel_group()
            )
        total_norm = total_norm.item() ** (1.0 / norm_type)
        clip_coef = max_norm / (total_norm + 1e-6)
        # Filter parameters with gradients.
        grads = [p.grad for p in parameters if p.grad is not None]
        if clip_coef < 1.0:
            grads_vec = fused_ops.TorchTensorVector(grads)
            multi_tensor_applier(fused_ops.multi_tensor_scale, dummy_overflow_buf, [grads_vec, grads_vec], clip_coef)

    return total_norm



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
"""
Creates an ONNX model with a single custom operator node that wraps an opaque model blob
meant to be executed by an external runtime. The model blob is serialized into the custom operator's
attributes.

Example:

python3 create_custom_op_wrapper.py --domain "test.domain"
        --custom_op_name my_custom_op
        --inputs "input0;FLOAT;1,10" "input1;FLOAT;1,10,10"
        --outputs "output0;STRING;10"
        --attribute_data xml googlenet-v1.xml
        --attribute_data bin googlenet-v1.bin
        -o test_model.onnx
"""

import argparse
import os
import sys
from dataclasses import dataclass

import onnx
from onnx import TensorProto, helper

IO_NAME_INDEX = 0
IO_ELEM_TYPE_INDEX = 1
IO_SHAPE_INDEX = 2

TENSOR_TYPE_MAP = {
    "UNDEFINED": TensorProto.UNDEFINED,
    "FLOAT": TensorProto.FLOAT,
    "UINT8": TensorProto.UINT8,
    "INT8": TensorProto.INT8,
    "UINT16": TensorProto.UINT16,
    "INT16": TensorProto.INT16,
    "INT32": TensorProto.INT32,
    "INT64": TensorProto.INT64,
    "STRING": TensorProto.STRING,
    "BOOL": TensorProto.BOOL,
    "FLOAT16": TensorProto.FLOAT16,
    "DOUBLE": TensorProto.DOUBLE,
    "UINT32": TensorProto.UINT32,
    "UINT64": TensorProto.UINT64,
    "COMPLEX64": TensorProto.COMPLEX64,
    "COMPLEX128": TensorProto.COMPLEX128,
    "BFLOAT16": TensorProto.BFLOAT16,
    "FLOAT8E4M3FN": TensorProto.FLOAT8E4M3FN,
    "FLOAT8E4M3FNUZ": TensorProto.FLOAT8E4M3FNUZ,
    "FLOAT8E5M2": TensorProto.FLOAT8E5M2,
    "FLOAT8E5M2FNUZ": TensorProto.FLOAT8E5M2FNUZ,
}


@dataclass
class IOInfo:
    """
    Class that represents the index, name, element type, and shape of an input or output.
    """

    index: int
    name: str
    elem_type: TensorProto.DataType
    shape: list[int | str] | None


def str_is_int(string: str) -> bool:
    try:
        int(string)
        return True
    except ValueError:
        return False


def parse_shape(shape_str: str) -> list[int | str] | None:
    try:
        shape = [int(s) if str_is_int(s) else s for s in shape_str.split(",")]
    except ValueError:
        shape = None

    return shape


class ParseIOInfoAction(argparse.Action):
    def __call__(self, parser, namespace, io_strs, opt_str):
        is_input = opt_str == "--inputs"
        io_meta_name = "input" if is_input else "output"
        ios = []

        for io_idx, io_str in enumerate(io_strs):
            comp_strs = []

            try:
                comp_strs = io_str.split(";")
            except ValueError:
                parser.error(f"{opt_str}: {io_meta_name} info must be separated by ';'")

            if len(comp_strs) != 3:
                parser.error(f"{opt_str}: {io_meta_name} info must have 3 components, but provided {len(comp_strs)}.")

            # Get io name
            io_name = comp_strs[IO_NAME_INDEX]

            # Get io element type
            io_elem_type = TENSOR_TYPE_MAP.get(comp_strs[IO_ELEM_TYPE_INDEX])
            if io_elem_type is None:
                type_options = ",".join(TENSOR_TYPE_MAP.keys())
                parser.error(
                    f"{opt_str}: invalid {io_meta_name} element type '{comp_strs[IO_ELEM_TYPE_INDEX]}'. "
                    f"Must be one of {type_options}."
                )

            # Get io shape
            io_shape = parse_shape(comp_strs[IO_SHAPE_INDEX])
            if io_shape is None:
                parser.error(
                    f"{opt_str}: invalid {io_meta_name} shape '{comp_strs[IO_SHAPE_INDEX]}'. "
                    "Expected comma-separated list of integers."
                )

            ios.append(
                IOInfo(
                    index=io_idx,
                    name=io_name,
                    elem_type=io_elem_type,
                    shape=io_shape,
                )
            )

        # Sort ios on index
        ios = sorted(ios, key=lambda elem: elem.index)

        setattr(namespace, self.dest, ios)


def parse_arguments() -> argparse.Namespace:
    io_metavar = '"<name>;<elem_type>;<shape>"'
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-n",
        "--custom_op_name",
        required=True,
        help="The custom operator's name.",
    )

    parser.add_argument(
        "-d",
        "--domain",
        required=True,
        help="The ONNX domain name used by the custom operator node.",
    )

    parser.add_argument(
        "--inputs",
        nargs="+",
        required=True,
        action=ParseIOInfoAction,
        help="List of strings specifying the name, element type, and shape of every input (in order). "
        'Ex: --inputs "input_0;FLOAT;1,10" "input_1;INT8;32"',
        metavar=io_metavar,
    )

    parser.add_argument(
        "--outputs",
        nargs="+",
        required=True,
        action=ParseIOInfoAction,
        help="List of strings specifying the name, element type, and shape of every output (in order). "
        'Ex: --outputs "output_0;FLOAT;1,10" "output_1;INT8;32"',
        metavar=io_metavar,
    )

    parser.add_argument(
        "-o",
        "--output_model",
        required=True,
        help="The name of the generated output model.",
    )

    parser.add_argument(
        "-a",
        "--attribute_data",
        nargs=2,
        required=False,
        action="append",
        help="The attribute name and file path of the data to serialize as a node attribute. "
        "Can be provided multiple times. Ex: -a model_xml ./model.xml -a model_bin ./model.bin",
    )

    parser.add_argument(
        "-v",
        "--opset_version",
        type=int,
        required=False,
        default=13,
        help="The Opset version.",
    )

    return parser.parse_args()


def get_attributes(attr_data_info: list[list[str]]):
    if not attr_data_info:
        return {}

    attrs = {}

    for info in attr_data_info:
        filepath = os.path.normpath(info[1])

        if not os.path.exists(filepath):
            print(f"[ERROR] attribute file '{info[1]}' does not exist.", file=sys.stderr)
            sys.exit(1)

        data = b""
        with open(filepath, "rb") as file_desc:
            data = file_desc.read()

        tensor = helper.make_tensor(name=info[0], data_type=TensorProto.UINT8, dims=[len(data)], vals=data, raw=True)

        attrs[info[0]] = tensor

    return attrs


def main():
    args = parse_arguments()

    inputs = []
    input_names = []
    for inp in args.inputs:
        inputs.append(helper.make_tensor_value_info(inp.name, inp.elem_type, inp.shape))
        input_names.append(inp.name)

    outputs = []
    output_names = []
    for out in args.outputs:
        outputs.append(helper.make_tensor_value_info(out.name, out.elem_type, out.shape))
        output_names.append(out.name)

    attrs = get_attributes(args.attribute_data)

    custom_op_node = helper.make_node(
        args.custom_op_name,
        name=args.custom_op_name + "_0",
        inputs=input_names,
        outputs=output_names,
        domain=args.domain,
        **attrs,
    )

    output_model_path = os.path.normpath(args.output_model)
    graph_name, model_ext = os.path.splitext(os.path.basename(output_model_path))

    if model_ext != ".onnx":
        print(
            f"[ERROR] Invalid output model name '{output_model_path}'. Must end in '.onnx'",
            file=sys.stderr,
        )
        sys.exit(1)

    graph_def = helper.make_graph([custom_op_node], graph_name, inputs, outputs)
    model_def = helper.make_model(graph_def, opset_imports=[helper.make_opsetid(args.domain, args.opset_version)])

    onnx.checker.check_model(model_def)
    onnx.save(model_def, output_model_path)

    print(f"[INFO] Saved output model to {output_model_path}")


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
from logging import getLogger

import numpy
from numpy import array_equal, ndarray
from onnx import NodeProto, TensorProto, helper, numpy_helper
from onnx import onnx_pb as onnx_proto
from onnx_model import OnnxModel

logger = getLogger(__name__)


class FusionUtils:
    def __init__(self, model: OnnxModel):
        self.model: OnnxModel = model

    def cast_graph_input_to_int32(self, input_name: str) -> tuple[bool, str]:
        graph_input = self.model.find_graph_input(input_name)
        if graph_input is not None and graph_input.type.tensor_type.elem_type != TensorProto.INT32:
            cast_output, cast_node = self.cast_input_to_int32(input_name)
            logger.debug(f"Casted graph input {input_name} to int32")
            return True, cast_output

        logger.debug(f"Did not cast graph input {input_name} to int32: found {graph_input is not None}")
        return False, input_name

    def cast_input(self, input_name: str, target_type="int32"):
        output_name = input_name + "_" + target_type

        if target_type == "int32":
            to_type = int(TensorProto.INT32)
        elif target_type == "float32":
            to_type = int(TensorProto.FLOAT)
        elif target_type == "float16":
            to_type = int(TensorProto.FLOAT16)
        else:
            raise ValueError("Invalid target_type: {target_type}")

        cast_node = self.add_cast_node(input_name, to_type, output_name)

        return output_name, cast_node

    def add_cast_node(
        self,
        input_name: str,
        to_type: int,
        output_name: str | None = None,
        output_name_to_node=None,
        graph_name: str | None = None,
    ):
        if output_name is None:
            output_name = input_name + f"_cast_to_{to_type}"

        # Avoid consequent Cast nodes.
        inputs = [input_name]
        if output_name_to_node is None:
            output_name_to_node = self.model.output_name_to_node()
        if input_name in output_name_to_node:
            parent_node = output_name_to_node[input_name]
            if parent_node and parent_node.op_type == "Cast":
                inputs = [parent_node.input[0]]

        cast_node = helper.make_node("Cast", inputs=inputs, outputs=[output_name])

        cast_node.attribute.extend([helper.make_attribute("to", to_type)])
        self.model.add_node(cast_node, graph_name=graph_name)

        return cast_node

    def cast_input_to_int32(self, input_name: str):
        return self.cast_input(input_name, "int32")

    def remove_cast_int32(self, input_name: str):
        input_name_to_nodes = self.model.input_name_to_nodes()
        nodes = input_name_to_nodes[input_name]
        for node in nodes:
            if node.op_type == "Cast":
                is_int32 = False
                for att in node.attribute:
                    if att.name == "to" and att.i == int(TensorProto.INT32):
                        is_int32 = True
                        break
                if is_int32:
                    output_name = node.output[0]
                    self.model.remove_node(node)
                    self.model.replace_input_of_all_nodes(output_name, input_name)

    @staticmethod
    def update_node_input(node, i, new_input_name, input_name_to_nodes):
        old_input_reference = 0
        if (node.input[i] in input_name_to_nodes) and node in input_name_to_nodes[node.input[i]]:
            input_name_to_nodes[node.input[i]].remove(node)
            old_input_reference = len(input_name_to_nodes[node.input[i]])

        node.input[i] = new_input_name

        if new_input_name in input_name_to_nodes:
            input_name_to_nodes[new_input_name].append(node)
        else:
            input_name_to_nodes[new_input_name] = [node]

        return old_input_reference

    @staticmethod
    def skip_parent(model: OnnxModel, node, parent_node, input_name_to_nodes, node_input_index=0, parent_input_index=0):
        """
        Before:
              (input)-->parent-->node-->(output)
        After:
              (input)-->parent-->
                |
                +----->node-->(output)

        This function returns a flag whether the parent node can be removed.
        """

        old_input_name = node.input[node_input_index]
        new_input_name = parent_node.input[parent_input_index]
        old_input_reference = FusionUtils.update_node_input(node, node_input_index, new_input_name, input_name_to_nodes)

        # We can remove the first Transpose if its output is not used (linked to graph output or other nodes) anymore.
        parent_can_be_removed = (old_input_reference == 0) and not model.find_graph_output(old_input_name)

        return parent_can_be_removed

    def get_squeeze_or_unsqueeze_axes(self, node: NodeProto) -> ndarray | None:
        assert node.op_type in ["Squeeze", "Unsqueeze"]

        # For opset >= 13, axes is an input instead of an attribute.
        if len(node.input) > 1:
            return self.model.get_constant_value(node.input[1])

        axes = None
        for attr in node.attribute:
            if attr.name == "axes":
                axes = helper.get_attribute_value(attr)
        return axes

    @staticmethod
    def check_node_attribute(node, attribute_name: str, expected_value, default_value=None):
        """Verify that a node has expected value for an attribute.

        Args:
            node (NodeProto): a node to check
            attribute_name (str): name of attribute
            expected_value (Any): expected value of the attribute
            default_value (Any, optional): default value if the attribute does not exist. Defaults to None.

        Returns:
            bool: whether the check is passed or not
        """
        value = default_value
        for attr in node.attribute:
            if attr.name == attribute_name:
                value = helper.get_attribute_value(attr)

        if isinstance(expected_value, list):
            return (isinstance(value, (ndarray, list))) and array_equal(expected_value, value, equal_nan=False)
        else:
            return value == expected_value

    @staticmethod
    def transpose_2d_int8_tensor(tensor: onnx_proto.TensorProto):
        """Transpose a 2-D INT8 TensorProto
        Args:
            tensor (TensorProto): tensor to be transposed
        Returns:
            tensor (TensorProto): transposed tensor
        """
        if not isinstance(tensor, onnx_proto.TensorProto):
            raise ValueError(f"Expected input type is an ONNX TensorProto but got {type(tensor)}")

        if len(tensor.dims) != 2 or tensor.data_type != onnx_proto.TensorProto.INT8:
            raise ValueError("Only INT8 2-D tensors can be transposed")

        if tensor.raw_data:
            int32_data = numpy.reshape(numpy.frombuffer(tensor.raw_data, dtype="int8"), tensor.dims)
            int32_transposed_data = numpy.transpose(int32_data, [1, 0])
            tensor.raw_data = int32_transposed_data.tobytes()

        else:
            raise ValueError("only raw buffer supported")

        return tensor

    @staticmethod
    def check_qdq_node_for_fusion(node: NodeProto, model: OnnxModel, allow_per_tensor_quantization_only=True):
        """Verify if a provided QuantizeLinear (Q) / DequantizeLinear (DQ) node is a good candidate for fusion.
           It is a good candidate for fusion if:
           (1) The Q/DQ node is for per-tensor quantization if allow_per_tensor_quantization_only is `True`
           (2) The Q/DQ node should have constant scale
           (3) The Q/DQ node should have a zero point of 0
        Args:
            node (NodeProto): a Q/DQ node to check
        Returns:
            bool: whether the check is passed or not
        """
        if node.op_type not in {"QuantizeLinear", "DequantizeLinear"}:
            logger.debug(f"Provided node is not a Q/DQ node. Op Type: {node.op_type}")

        scale = model.get_constant_value(node.input[1])

        # Scale is not constant
        if scale is None:
            return False

        # Not per-tensor quantization
        scale_has_single_element = scale.ndim == 0 or (scale.ndim == 1 and scale.shape[0] == 1)
        if allow_per_tensor_quantization_only and not scale_has_single_element:
            return False

        # If the Q/DQ node has no zero point input, it is assumed to be 0 (per ONNX spec)
        if len(node.input) == 2:
            return True

        # Zero point should be constant and should have a value of 0
        zero_point = model.get_constant_value(node.input[2])

        # Zero point and scale should have same number of dims
        if scale.ndim != zero_point.ndim:
            return False

        # Zero point is not constant or zero point is not zero
        if zero_point is None:
            return False

        return numpy.all(zero_point == 0)

    def check_node_input_value(self, node, input_index: int, expected_value):
        """Verify that a node has expected input value

        Args:
            node (NodeProto): a node to check
            input_index (int): index of its input to be verified
            expected_value (Any): expected value of the input

        Returns:
            bool: whether the check is passed or not
        """
        assert len(node.input) > input_index

        value = self.model.get_constant_value(node.input[input_index])

        if isinstance(expected_value, list):
            return (isinstance(value, (ndarray, list))) and array_equal(expected_value, value, equal_nan=False)
        else:
            return value == expected_value

    def remove_identity_nodes(self):
        """Remove Identity nodes, except those right before graph output."""
        nodes_to_remove = []
        graph_output_names = self.model.get_graphs_output_names()
        for node in self.model.nodes():
            if node.op_type == "Identity":
                if node.output[0] not in graph_output_names:
                    self.model.replace_input_of_all_nodes(node.output[0], node.input[0])
                    nodes_to_remove.append(node)

        if nodes_to_remove:
            self.model.remove_nodes(nodes_to_remove)
            logger.info(f"Removed {len(nodes_to_remove)} Identity nodes")

    def remove_cascaded_cast_nodes(self):
        self.model.remove_cascaded_cast_nodes()

    def remove_useless_cast_nodes(self):
        self.model.remove_useless_cast_nodes()

    def remove_useless_reshape_nodes(self):
        """Remove reshape node that is not needed based on symbolic shape inference: input and output has same shape"""
        shape_infer = self.model.infer_runtime_shape(update=True)
        if shape_infer is None:
            return

        nodes_to_remove = []
        for node in self.model.nodes():
            if node.op_type == "Reshape":
                input_shape = shape_infer.get_edge_shape(node.input[0])
                output_shape = shape_infer.get_edge_shape(node.output[0])
                if input_shape and output_shape and input_shape == output_shape:
                    logger.info(
                        f"Remove reshape node {node.name} since its input shape is same as output: {input_shape}"
                    )
                    nodes_to_remove.append(node)

        if nodes_to_remove:
            graph_input_names = set(self.model.get_graphs_input_names())
            graph_output_names = set(self.model.get_graphs_output_names())
            for node in nodes_to_remove:
                if bool(set(node.output) & graph_output_names):
                    if (
                        not bool(set(node.input) & graph_input_names)
                        and len(self.model.input_name_to_nodes()[node.input[0]]) == 1  # parent has only one child
                    ):
                        self.model.replace_output_of_all_nodes(node.input[0], node.output[0])
                    else:
                        continue
                else:
                    self.model.replace_input_of_all_nodes(node.output[0], node.input[0])
                self.model.remove_node(node)


class NumpyHelper:
    @staticmethod
    def to_array(tensor: TensorProto, fill_zeros: bool = False) -> ndarray:
        # When weights are in external data format but not presented, we can still test the optimizer with two changes:
        # (1) set fill_zeros = True  (2) change load_external_data=False in optimizer.py
        if fill_zeros:
            from onnx import mapping

            return ndarray(
                shape=tensor.dims,
                dtype=mapping.TENSOR_TYPE_TO_NP_TYPE[tensor.data_type],
            )

        return numpy_helper.to_array(tensor)



# !/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from __future__ import annotations

import argparse
import io
import re
import shutil
import sys
import typing
from pathlib import Path

import op_registration_utils
from logger import get_logger

# directory containing the reduced op files, relative to the build directory
OP_REDUCTION_DIR = "op_reduction.generated"

# add the path to tools/python so we can import the config parsing and type reduction processing
SCRIPT_DIR = Path(__file__).parent.resolve()
ORT_ROOT = SCRIPT_DIR.parents[1]
sys.path.append(str(ORT_ROOT / "tools" / "python"))

from util import parse_config  # noqa: E402
from util.ort_format_model.operator_type_usage_processors import OpTypeImplFilterInterface  # noqa: E402

log = get_logger("reduce_op_kernels")


def _adapt_filters_for_extended_minimal_build(
    base_required_ops: dict | None, base_op_type_impl_filter: OpTypeImplFilterInterface | None
):
    """
    Adapts the values returned by parse_config() for an extended minimal build or higher.
    In particular:
    - Includes ONNX ops needed by layout transformation
    - Includes MS ops needed by NHWC optimizer
    """
    # graph transformations in an extended minimal build require certain ops to be available
    extended_minimal_build_required_op_ids = set()  # set of (domain, optype, opset)
    with open(
        ORT_ROOT / "onnxruntime/core/optimizer/layout_transformation/layout_transformation_potentially_added_ops.h",
    ) as f:
        region_boundary_pattern = re.compile(r"@@region_(begin|end)\(extended_minimal_build_required_kernels\)@@")
        op_id_pattern = re.compile(
            r'OpIdentifierWithStringViews{(?P<domain>\w+),\s+"(?P<optype>\w+)",\s+(?P<opset>\d+)}'
        )
        in_region = False
        for line in f:
            region_boundary_match = region_boundary_pattern.search(line)
            if region_boundary_match:
                in_region = region_boundary_match.group(1) == "begin"
                continue

            if not in_region:
                continue

            op_id_match = op_id_pattern.search(line)
            if op_id_match:
                domain = op_registration_utils.map_ort_constant_to_domain(
                    op_id_match.group("domain"), allow_unknown_constant=False
                )
                optype = op_id_match.group("optype")
                opset = int(op_id_match.group("opset"))
                extended_minimal_build_required_op_ids.add((domain, optype, opset))

    adapted_required_ops = None
    if base_required_ops is not None:
        adapted_required_ops = base_required_ops.copy()
        for domain, optype, opset in extended_minimal_build_required_op_ids:
            adapted_required_ops.setdefault(domain, {}).setdefault(opset, set()).add(optype)

    adapted_op_type_impl_filter = None
    if base_op_type_impl_filter is not None:

        class _AdaptedFilter(OpTypeImplFilterInterface):
            def __init__(
                self,
                filter_to_adapt: OpTypeImplFilterInterface,
                required_domain_and_optypes: set[tuple[str, str]],
            ):
                self.filter_to_adapt = filter_to_adapt
                self.required_domain_and_optypes = required_domain_and_optypes

            def is_typed_registration_needed(self, domain: str, optype: str, type_registration_str: str):
                # Always require registration for ops in self.required_domain_and_optypes.
                if (domain, optype) in self.required_domain_and_optypes:
                    return True
                return self.filter_to_adapt.is_typed_registration_needed(domain, optype, type_registration_str)

            def get_cpp_entries(self):
                # The required types for ops in self.required_optypes must be specified in the C++ implementation.
                # Doing that also accounts for globally allowed types.
                # We don't need to do anything special with the allowed type overrides here.
                return self.filter_to_adapt.get_cpp_entries()

        adapted_op_type_impl_filter = _AdaptedFilter(
            base_op_type_impl_filter,
            {(domain, optype) for (domain, optype, opset) in extended_minimal_build_required_op_ids},
        )

    return (adapted_required_ops, adapted_op_type_impl_filter)


class _ExcludingRegistrationProcessor(op_registration_utils.RegistrationProcessor):
    """Registration processor that excludes registrations and writes the result to an output file."""

    def __init__(
        self,
        required_ops: dict | None,
        op_type_impl_filter: OpTypeImplFilterInterface | None,
        output_file: io.TextIOWrapper,
    ):
        self._required_ops = required_ops
        self._op_type_impl_filter = op_type_impl_filter
        self._output_file = output_file

    def _is_op_required(self, domain: str, operator: str, start_version: int, end_version: int | None) -> bool:
        """See if an op is required."""
        if self._required_ops is None:
            return True

        if domain not in self._required_ops:
            return False

        for opset in self._required_ops[domain]:
            if opset >= start_version and (end_version is None or opset <= end_version):
                if operator in self._required_ops[domain][opset]:
                    return True

        return False

    def process_registration(
        self,
        lines: list[str],
        constant_for_domain: str,
        operator: str,
        start_version: int,
        end_version: int | None = None,
        type: str | None = None,
    ):
        registration_identifier = "{}:{}({}){}".format(
            constant_for_domain, operator, start_version, f"<{type}>" if type else ""
        )

        # convert from the ORT constant name to the domain string used in the config
        domain = op_registration_utils.map_ort_constant_to_domain(constant_for_domain, allow_unknown_constant=False)

        exclude = False
        reason = ""

        if domain is not None:
            if not self._is_op_required(domain, operator, start_version, end_version):
                exclude = True
                reason = "Entire op is not required."

            if not exclude and type is not None and self._op_type_impl_filter is not None:
                if not self._op_type_impl_filter.is_typed_registration_needed(domain, operator, type):
                    exclude = True
                    reason = "Specific typed registration is not required."
        else:
            log.warning(f"Keeping {registration_identifier} registration from unknown domain: {constant_for_domain}")

        if exclude:
            log.info(f"Disabling {registration_identifier} registration: {reason}")
            for line in lines:
                self._output_file.write("// " + line)

            # edge case of last entry in table where we still need the terminating }; to not be commented out
            if lines[-1].rstrip().endswith("};"):
                self._output_file.write("};\n")
        else:
            for line in lines:
                self._output_file.write(line)

    def process_other_line(self, line):
        self._output_file.write(line)

    def ok(self):
        return True


def _get_op_reduction_root(build_dir: Path):
    """
    Return the op reduction root directory which is a subdirectory of `build_dir`.
    """
    return Path(build_dir, OP_REDUCTION_DIR)


def _get_op_reduction_file_path(ort_root: Path, build_dir: Path, original_path: Path):
    """
    Return the op reduction file path corresponding to `original_path`.
    Op reduction files are in the op reduction root but otherwise share the same components of `original_path`
    relative to `ort_root`.
    """
    return _get_op_reduction_root(build_dir) / original_path.relative_to(ort_root)


def _generate_provider_registrations(
    ort_root: Path,
    build_dir: Path,
    use_cuda: bool,
    required_ops: dict | None,
    op_type_impl_filter: OpTypeImplFilterInterface | None,
):
    """Generate provider registration files."""
    kernel_registration_files = [
        Path(f) for f in op_registration_utils.get_kernel_registration_files(str(ort_root), use_cuda)
    ]

    for kernel_registration_file in kernel_registration_files:
        if not kernel_registration_file.is_file():
            raise ValueError(f"Kernel registration file does not exist: {kernel_registration_file}")

        log.info(f"Processing {kernel_registration_file}")

        reduced_path = _get_op_reduction_file_path(ort_root, build_dir, kernel_registration_file)

        reduced_path.parent.mkdir(parents=True, exist_ok=True)

        # read from original and create the reduced kernel def file with commented out lines for any kernels that are
        # not required
        with open(reduced_path, "w") as file_to_write:
            processor = _ExcludingRegistrationProcessor(required_ops, op_type_impl_filter, file_to_write)

            op_registration_utils.process_kernel_registration_file(kernel_registration_file, processor)

            if not processor.ok():
                # error should have already been logged so just exit
                sys.exit(-1)


def _generate_type_control_overrides(ort_root: Path, build_dir: Path, cpp_lines: typing.Sequence[str]):
    """
    Generate type control overrides. Insert applicable C++ code to specify operator type requirements.
    :param ort_root: Root of the ONNX Runtime repository
    :param build_dir: Path to the build directory
    :param cpp_lines: The C++ code to insert
    """
    src = Path(ort_root, "onnxruntime", "core", "providers", "op_kernel_type_control_overrides.inc")

    if not src.is_file():
        raise ValueError(f"Op kernel type control overrides file does not exist: {src}")

    # create a copy of op_kernel_type_control_overrides.inc
    target = _get_op_reduction_file_path(ort_root, build_dir, src)

    target.parent.mkdir(parents=True, exist_ok=True)

    shutil.copyfile(src, target)

    if cpp_lines:
        # find the insertion block and replace any existing content in it
        inserted = False
        with open(src) as input, open(target, "w") as output:
            inside_insertion_block = False
            for line in input:
                if "@@insertion_point_begin(allowed_types)@@" in line:
                    inside_insertion_block = True
                    output.write(line)
                    [output.write(f"{code_line}\n") for code_line in cpp_lines]
                    inserted = True
                    continue
                elif inside_insertion_block:
                    if "@@insertion_point_end(allowed_types)@@" in line:
                        inside_insertion_block = False
                    else:
                        # we ignore any old lines within the insertion block
                        continue

                output.write(line)

        if not inserted:
            raise RuntimeError(f"Insertion point was not found in {target}")


def reduce_ops(
    config_path: str,
    build_dir: str,
    enable_type_reduction: bool,
    use_cuda: bool,
    is_extended_minimal_build_or_higher: bool,
):
    """
    Reduce op kernel implementations.
    :param config_path: Path to configuration file that specifies the ops to include
    :param build_dir: Path to the build directory. The op reduction files will be generated under the build directory.
    :param enable_type_reduction: Whether per operator type reduction is enabled
    :param use_cuda: Whether to reduce op kernels for the CUDA provider
    :param is_extended_minimal_build_or_higher: Whether this build has at least the features of an extended minimal
                                                build enabled.
    """
    build_dir_path = Path(build_dir).resolve()
    build_dir_path.mkdir(parents=True, exist_ok=True)

    required_ops, op_type_impl_filter = parse_config(config_path, enable_type_reduction)
    if is_extended_minimal_build_or_higher:
        required_ops, op_type_impl_filter = _adapt_filters_for_extended_minimal_build(required_ops, op_type_impl_filter)

    # delete any existing generated files first
    op_reduction_root = _get_op_reduction_root(build_dir_path)
    if op_reduction_root.is_dir():
        log.info(f"Deleting existing op reduction file root directory: {op_reduction_root}")
        shutil.rmtree(op_reduction_root)

    _generate_provider_registrations(ORT_ROOT, build_dir_path, use_cuda, required_ops, op_type_impl_filter)

    type_control_cpp_code = op_type_impl_filter.get_cpp_entries() if op_type_impl_filter is not None else []
    _generate_type_control_overrides(ORT_ROOT, build_dir_path, type_control_cpp_code)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Reduces operator kernel implementations in ONNX Runtime. "
        "Entire op implementations or op implementations for specific types may be pruned."
    )

    parser.add_argument(
        "config_path",
        type=str,
        help="Path to configuration file. "
        "Create with <ORT root>/tools/python/create_reduced_build_config.py and edit if needed. "
        "See https://onnxruntime.ai/docs/reference/operators/reduced-operator-config-file.html for more "
        "information.",
    )

    parser.add_argument(
        "--cmake_build_dir",
        type=str,
        required=True,
        help="Path to the build directory. The op reduction files will be generated under the build directory.",
    )

    parser.add_argument(
        "--is_extended_minimal_build_or_higher",
        action="store_true",
        help="Whether this build has at least the features of an extended minimal build enabled.",
    )

    parser.add_argument(
        "--enable_type_reduction", action="store_true", help="Whether per operator type reduction is enabled."
    )

    parser.add_argument("--use_cuda", action="store_true", help="Whether to reduce op kernels for the CUDA provider.")

    args = parser.parse_args()

    reduce_ops(
        config_path=args.config_path,
        build_dir=args.cmake_build_dir,
        enable_type_reduction=args.enable_type_reduction,
        use_cuda=args.use_cuda,
        is_extended_minimal_build_or_higher=args.is_extended_minimal_build_or_higher,
    )



"""
Run this script to recreate the original onnx model.
Example usage:
python packed_attention_fp16.model.py out_model_path.onnx
"""

import sys

import numpy as np
import onnx
from onnx import TensorProto, helper, numpy_helper


def clear_field(proto, field):
    proto.ClearField(field)
    return proto


def order_repeated_field(repeated_proto, key_name, order):
    order = list(order)
    repeated_proto.sort(key=lambda x: order.index(getattr(x, key_name)))


def make_node(op_type, inputs, outputs, name=None, doc_string=None, domain=None, **kwargs):
    node = helper.make_node(op_type, inputs, outputs, name, doc_string, domain, **kwargs)
    if not doc_string:
        node.doc_string = ""
    order_repeated_field(node.attribute, "name", kwargs.keys())
    return node


def make_graph(*args, doc_string=None, **kwargs):
    graph = helper.make_graph(*args, doc_string=doc_string, **kwargs)
    if not doc_string:
        graph.doc_string = ""
    return graph


model = helper.make_model(
    opset_imports=[
        clear_field(helper.make_operatorsetid("", 12), "domain"),
        helper.make_operatorsetid("com.microsoft", 1),
    ],
    ir_version=7,
    producer_name="onnxruntime.transformers",
    producer_version="1.15.0",
    graph=make_graph(
        name="torch-jit-export",
        inputs=[
            helper.make_tensor_value_info("input", TensorProto.FLOAT16, shape=["token_count", "hidden_size"]),
            helper.make_tensor_value_info("weight", TensorProto.FLOAT16, shape=["hidden_size", "hidden_size_x_3"]),
            helper.make_tensor_value_info("bias", TensorProto.FLOAT16, shape=["hidden_size_x_3"]),
            helper.make_tensor_value_info("token_offset", TensorProto.INT32, shape=["batch_size", "seq_len"]),
            helper.make_tensor_value_info("cumulative_sequence_length", TensorProto.INT32, shape=["batch_size_plus_1"]),
            helper.make_tensor_value_info(
                "rbp", TensorProto.FLOAT16, shape=["batch_size", "num_heads", "seq_len", "seq_len"]
            ),
        ],
        outputs=[helper.make_tensor_value_info("output", TensorProto.FLOAT16, shape=["token_count", "hidden_size"])],
        nodes=[
            make_node(
                "Constant",
                inputs=[],
                outputs=["constant_0"],
                name="constant_0",
                value=numpy_helper.from_array(np.array([0], dtype="int64"), name=""),
            ),
            make_node(
                "Constant",
                inputs=[],
                outputs=["constant_1"],
                name="Constant_1",
                value=numpy_helper.from_array(np.array([1], dtype="int64"), name=""),
            ),
            make_node(
                "Constant",
                inputs=[],
                outputs=["constant_last"],
                name="constant_last",
                value=numpy_helper.from_array(np.array([-1], dtype="int64"), name=""),
            ),
            make_node(
                "Constant",
                inputs=[],
                outputs=["constant_max"],
                name="Constant_max",
                value=numpy_helper.from_array(np.array([9223372036854775807], dtype="int64"), name=""),
            ),
            make_node(
                "Slice",
                inputs=["cumulative_sequence_length", "constant_0", "constant_last"],
                outputs=["start"],
                name="slice_start",
            ),
            make_node(
                "Slice",
                inputs=["cumulative_sequence_length", "constant_1", "constant_max"],
                outputs=["end"],
                name="slice_end",
            ),
            make_node("Sub", inputs=["end", "start"], outputs=["mask_idx"]),
            make_node(
                "RestorePadding",
                inputs=["input", "token_offset"],
                outputs=["restore_padding_output"],
                name="RestorePadding_1",
                domain="com.microsoft",
            ),
            make_node(
                "Attention",
                inputs=["restore_padding_output", "weight", "bias", "mask_idx", "", "rbp"],
                outputs=["attention_outputs"],
                name="Attention_0",
                domain="com.microsoft",
                num_heads=12,
                mask_filter_value=-3.4028234663852886e38,
            ),
            make_node(
                "RemovePadding",
                inputs=["attention_outputs", "mask_idx"],
                outputs=["output", "remove_padding_token_offset", "cumulated_seq_len", "max_seq_len"],
                name="RemovePadding_0",
                domain="com.microsoft",
            ),
        ],
    ),
)

if __name__ == "__main__" and len(sys.argv) == 2:
    _, out_path = sys.argv
    onnx.save(model, out_path)



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import pathlib
import subprocess

SCRIPT_DIR = pathlib.Path(__file__).parent.resolve()


def generate_cpp(flatc: pathlib.Path, schema_path: pathlib.Path):
    # run flatc to generate C++ code
    cmd = [str(flatc), "--cpp", "--scoped-enums", "--filename-suffix", ".fbs", str(schema_path)]
    subprocess.run(cmd, check=True, cwd=SCRIPT_DIR)


def main():
    parser = argparse.ArgumentParser(
        description="Generate language bindings for the ORT flatbuffers schema.",
        usage="Provide the path to the flatbuffers flatc executable. "
        "Script can be executed from anywhere but must be located in its original "
        "directory in the ONNX Runtime enlistment.",
    )

    parser.add_argument(
        "-f",
        "--flatc",
        required=True,
        type=pathlib.Path,
        help="Path to flatbuffers flatc executable. "
        "Can be found in the build directory under _deps/flatbuffers-build/<config>/",
    )

    all_languages = ["cpp"]
    parser.add_argument(
        "-l",
        "--language",
        action="append",
        dest="languages",
        choices=all_languages,
        help="Specify which language bindings to generate.",
    )

    args = parser.parse_args()
    languages = args.languages if args.languages is not None else all_languages
    flatc = args.flatc.resolve(strict=True)
    schema_path = SCRIPT_DIR / "adapter_schema.fbs"

    if "cpp" in languages:
        generate_cpp(flatc, schema_path)


if __name__ == "__main__":
    main()



import onnx
from onnx import OperatorSetIdProto, TensorProto, helper

X1 = helper.make_tensor_value_info("x1", TensorProto.INT64, [4, 4])
X2 = helper.make_tensor_value_info("x2", TensorProto.INT64, [4, 4])
Y1 = helper.make_tensor_value_info("output1", TensorProto.INT64, [4, 4])
Y2 = helper.make_tensor_value_info("output2", TensorProto.INT64, [4, 4])

add1 = helper.make_node("Add", ["x1", "x2"], ["add1"], name="add1")
add2 = helper.make_node("Add", ["x1", "x2"], ["add2"], name="add2")
id1 = helper.make_node("Identity", ["add1"], ["output1"], name="id1")
id2 = helper.make_node("Identity", ["add2"], ["output2"], name="id2")

# Create the graph (GraphProto)
graph_def = helper.make_graph([add1, add2, id1, id2], "identity_elimination_model", [X1, X2], [Y1, Y2])

opsets = []
onnxdomain = OperatorSetIdProto()
onnxdomain.version = 12
onnxdomain.domain = ""  # The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
opsets.append(onnxdomain)

msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"

opsets.append(msdomain)
kwargs = {}
kwargs["opset_imports"] = opsets

# Create the model (ModelProto)
model_def = helper.make_model(graph_def, producer_name="onnx-example", **kwargs)
onnx.save(model_def, "id-elim.onnx")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import copy
import json
import os
import random
import uuid

import _test_helpers
import onnx
import pytest
import torch
from onnx import TensorProto, helper
from torch._C import _from_dlpack
from torch.utils.dlpack import to_dlpack

from onnxruntime.training.ort_triton import call_triton_by_name, call_triton_by_onnx
from onnxruntime.training.ortmodule import DebugOptions, ORTModule

pytest.importorskip("triton")

DEVICE = "cuda"


@pytest.fixture(scope="session", autouse=True)
def run_before_test_session(request):
    def insert_use_triton_env():
        os.environ["ORTMODULE_USE_TRITON"] = "1"

    def remove_use_triton_env():
        del os.environ["ORTMODULE_USE_TRITON"]

    insert_use_triton_env()
    request.addfinalizer(remove_use_triton_env)


def _onnx_dtype_to_torch_dtype(onnx_dtype):
    if onnx_dtype == TensorProto.FLOAT:
        return torch.float32
    elif onnx_dtype == TensorProto.FLOAT16:
        return torch.float16
    else:
        raise RuntimeError("Unsupported ONNX dtype")


def _torch_add(input1, input2):
    return input1 + input2


def _torch_sub(input1, input2):
    return input1 - input2


def _torch_mul(input1, input2):
    return input1 * input2


def _torch_div(input1, input2):
    return input1 / input2


def _torch_pow(input, exponent):
    return torch.pow(input, exponent)


def _torch_sqrt(input):
    return torch.sqrt(input)


def _torch_exp(input):
    return torch.exp(input)


def _torch_cast(input, **kwargs):
    return input.to(_onnx_dtype_to_torch_dtype(kwargs.get("to")))


def _torch_where(condition, x, y):
    return torch.where(condition, x, y)


def _torch_sum(*inputs):
    result = inputs[0]
    for input in inputs[1:]:
        result += input
    return result


def _troch_dropout_gard(dy, mask, **kwargs):
    ratio = kwargs.get("ratio", 0.5)
    return torch.where(mask, dy / (1.0 - ratio), 0.0)


def _torch_softmax(input, **kwargs):
    axis = kwargs.get("axis", -1)
    return torch.softmax(input, axis)


def _torch_reduce(input, func, **kwargs):
    rank = len(input.shape)
    axes = kwargs.get("axes", list(range(rank)))
    keepdims = kwargs.get("keepdims", True)
    axes = [axis if axis >= 0 else rank + axis for axis in axes]
    axes.sort(reverse=True)
    result = input
    for axis in axes:
        result = func(result, dim=axis, keepdim=keepdims)
        if func == torch.max or func == torch.min:
            result = result[0]
    return result


def _torch_reduce_sum(input, **kwargs):
    return _torch_reduce(input, torch.sum, **kwargs)


def _torch_reduce_mean(input, **kwargs):
    return _torch_reduce(input, torch.mean, **kwargs)


def _torch_reduce_max(input, **kwargs):
    return _torch_reduce(input, torch.max, **kwargs)


def _torch_reduce_min(input, **kwargs):
    return _torch_reduce(input, torch.min, **kwargs)


def _torch_layer_norm(input, weight, bias, **kwargs):
    rank = len(input.shape)
    axis = kwargs.get("axis", -1)
    if axis < 0:
        axis += rank
    normalized_shape = input.shape[axis:]
    return torch.nn.functional.layer_norm(input, normalized_shape, weight, bias)


def _torch_gelu(input):
    return torch.nn.functional.gelu(input)


def _torch_quick_gelu(input, **kwargs):
    alpha = kwargs.get("alpha", 1.702)
    return input * torch.sigmoid(input * alpha)


def _torch_gelu_grad(dy, x):
    alpha = 0.70710678118654752440
    beta = 1.12837916709551257390 * 0.70710678118654752440 * 0.5
    cdf = 0.5 * (1 + torch.erf(x * alpha))
    pdf = beta * torch.exp(x * x * -0.5)
    return dy * (cdf + x * pdf)


def _torch_quick_gelu_grad(dy, x, **kwargs):
    alpha = kwargs.get("alpha", 1.702)
    sigmoid = torch.sigmoid(x * alpha)
    return dy * sigmoid * (1.0 + x * alpha * (1.0 - sigmoid))


class TorchFuncExecutor:
    _TORCH_FUNC_MAP = {  # noqa: RUF012
        "Add": _torch_add,
        "Sub": _torch_sub,
        "Mul": _torch_mul,
        "Div": _torch_div,
        "Pow": _torch_pow,
        "Sqrt": _torch_sqrt,
        "Exp": _torch_exp,
        "Cast": _torch_cast,
        "Where": _torch_where,
        "Sum": _torch_sum,
        "DropoutGrad": _troch_dropout_gard,
        "ReduceSum": _torch_reduce_sum,
        "ReduceMean": _torch_reduce_mean,
        "ReduceMax": _torch_reduce_max,
        "ReduceMin": _torch_reduce_min,
        "Softmax": _torch_softmax,
        "LayerNormalization": _torch_layer_norm,
        "Gelu": _torch_gelu,
        "QuickGelu": _torch_quick_gelu,
        "GeluGrad": _torch_gelu_grad,
        "QuickGeluGrad": _torch_quick_gelu_grad,
    }

    @classmethod
    def run(cls, op_type, *torch_tensors, **kwargs):
        if op_type not in cls._TORCH_FUNC_MAP:
            raise NotImplementedError(f"Unsupported op type: {op_type}")
        return cls._TORCH_FUNC_MAP[op_type](*torch_tensors, **kwargs)


def _run_op_test(op_type, onnx_dtype, create_model_func, gen_inputs_func, **kwargs):
    rtol = kwargs.get("rtol", 1e-03 if onnx_dtype == TensorProto.FLOAT16 else 1e-04)
    atol = kwargs.get("atol", 1e-03 if onnx_dtype == TensorProto.FLOAT16 else 1e-05)
    pt_inputs = gen_inputs_func(_onnx_dtype_to_torch_dtype(onnx_dtype))
    ort_inputs = copy.deepcopy(pt_inputs)
    ort_inputs = [tensor.to(torch.uint8) if tensor.dtype == torch.bool else tensor for tensor in ort_inputs]
    if "::" in op_type:
        _, op_type = op_type.split("::")
    pt_outputs = TorchFuncExecutor.run(op_type, *pt_inputs, **kwargs)
    model_str = create_model_func(op_type, onnx_dtype, **kwargs).SerializeToString()
    unique_id = uuid.uuid1().int >> 64
    ort_outputs = call_triton_by_onnx(unique_id, model_str, *[to_dlpack(tensor) for tensor in ort_inputs])
    if isinstance(pt_outputs, tuple):
        assert isinstance(ort_outputs, tuple)
        assert len(pt_outputs) == len(ort_outputs)
        for pt_output, ort_output in zip(pt_outputs, ort_outputs, strict=False):
            _test_helpers.assert_values_are_close(pt_output, _from_dlpack(ort_output), rtol=rtol, atol=atol)
    else:
        _test_helpers.assert_values_are_close(pt_outputs, _from_dlpack(ort_outputs), rtol=rtol, atol=atol)


def _run_step(model, *tensors):
    prediction = model(*tensors)
    loss = prediction.sum()
    loss.backward()
    return prediction


def _run_module_test(module_cls, dtype, gen_inputs_func, triton_op_count, **kwargs):
    pt_model = module_cls().to(DEVICE).to(dtype)
    ort_model = ORTModule(copy.deepcopy(pt_model), DebugOptions(save_onnx=True, onnx_prefix="triton_model"))
    rtol = kwargs.get("rtol", 1e-03 if dtype == torch.float16 else 1e-04)
    atol = kwargs.get("atol", 1e-03 if dtype == torch.float16 else 1e-05)
    for _ in range(10):
        pt_inputs = gen_inputs_func(dtype)
        ort_inputs = copy.deepcopy(pt_inputs)
        pt_output = _run_step(pt_model, *pt_inputs)
        ort_output = _run_step(ort_model, *ort_inputs)
        _test_helpers.assert_values_are_close(pt_output, ort_output, rtol=rtol, atol=atol)
        _test_helpers.assert_gradients_match_and_reset_gradient(pt_model, ort_model, rtol=rtol, atol=atol)
        for idx, pt_input in enumerate(pt_inputs):
            if pt_input.requires_grad:
                _test_helpers.assert_values_are_close(pt_input.grad, ort_inputs[idx].grad, rtol=rtol, atol=atol)

    assert os.path.exists(os.path.join(os.getcwd(), "triton_model_torch_exported_training.onnx"))
    assert os.path.exists(os.path.join(os.getcwd(), "triton_model_optimized_training.onnx"))
    assert os.path.exists(os.path.join(os.getcwd(), "triton_model_optimized_pre_grad_training.onnx"))
    assert os.path.exists(os.path.join(os.getcwd(), "triton_model_execution_model_training.onnx"))
    model = onnx.load(os.path.join(os.getcwd(), "triton_model_execution_model_training.onnx"))
    actual_triton_op_count = 0
    for node in model.graph.node:
        if node.op_type == "TritonOp":
            actual_triton_op_count += 1
    assert actual_triton_op_count == triton_op_count
    os.remove(os.path.join(os.getcwd(), "triton_model_torch_exported_training.onnx"))
    os.remove(os.path.join(os.getcwd(), "triton_model_optimized_training.onnx"))
    os.remove(os.path.join(os.getcwd(), "triton_model_optimized_pre_grad_training.onnx"))
    os.remove(os.path.join(os.getcwd(), "triton_model_execution_model_training.onnx"))


def _run_tunable_op_test(module_cls, dtype, gen_inputs_func, tunable_op, impl_count, **kwargs):
    os.environ["ORTMODULE_ENABLE_TUNING"] = "1"
    os.environ["ORTMODULE_TUNING_RESULTS_PATH"] = "./"
    pt_model = module_cls().to(DEVICE).to(dtype)
    ort_model = ORTModule(copy.deepcopy(pt_model))
    rtol = kwargs.get("rtol", 1e-03 if dtype == torch.float16 else 1e-04)
    atol = kwargs.get("atol", 1e-03 if dtype == torch.float16 else 1e-05)
    for _ in range(5):
        pt_inputs = gen_inputs_func(dtype)
        ort_inputs = copy.deepcopy(pt_inputs)
        pt_output = _run_step(pt_model, *pt_inputs)
        ort_output = _run_step(ort_model, *ort_inputs)
        _test_helpers.assert_values_are_close(pt_output, ort_output, rtol=rtol, atol=atol)
        _test_helpers.assert_gradients_match_and_reset_gradient(pt_model, ort_model, rtol=rtol, atol=atol)
    tunable_results_file = os.path.join(os.getcwd(), "tuning_results_training.json")
    assert os.path.exists(tunable_results_file)
    with open(tunable_results_file, encoding="UTF-8") as f:
        tunable_results = json.load(f)
    assert tunable_op in str(tunable_results)
    del os.environ["ORTMODULE_ENABLE_TUNING"]
    for i in range(impl_count - 1):
        new_tunable_results = copy.deepcopy(tunable_results)
        for k, v in new_tunable_results[0]["results"].items():
            if tunable_op in k:
                for param, impl in v.items():
                    v[param] = (impl + 1 + i) % impl_count
        with open(tunable_results_file, "w", encoding="UTF-8") as f:
            json.dump(new_tunable_results, f)
        ort_model = ORTModule(copy.deepcopy(pt_model))
        for _ in range(5):
            pt_inputs = gen_inputs_func(dtype)
            ort_inputs = copy.deepcopy(pt_inputs)
            pt_output = _run_step(pt_model, *pt_inputs)
            ort_output = _run_step(ort_model, *ort_inputs)
            _test_helpers.assert_values_are_close(pt_output, ort_output, rtol=rtol, atol=atol)
            _test_helpers.assert_gradients_match_and_reset_gradient(pt_model, ort_model, rtol=rtol, atol=atol)
    os.remove(tunable_results_file)
    del os.environ["ORTMODULE_TUNING_RESULTS_PATH"]


@pytest.mark.parametrize(
    "op",
    [
        ("Add", {}),
        ("Sub", {}),
        ("Mul", {}),
        ("Div", {}),
        ("com.microsoft::GeluGrad", {}),
        ("com.microsoft::QuickGeluGrad", {}),
        ("com.microsoft::QuickGeluGrad", {"alpha": 1.0}),
    ],
)
@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shapes", [([1024, 2], [1024, 2]), ([2, 3, 3, 3], [3, 1, 3]), ([2049], [1])])
def test_binary_elementwise_op(op, onnx_dtype, input_shapes):
    def _create_model(op_type, onnx_dtype, **kwargs):
        domain = ""
        if "::" in op_type:
            domain, op_type = op_type.split("::")
        graph = helper.make_graph(
            [helper.make_node(op_type, ["X", "Y"], ["Z"], name="test", domain=domain, **kwargs)],
            "test",
            [
                helper.make_tensor_value_info("X", onnx_dtype, None),
                helper.make_tensor_value_info("Y", onnx_dtype, None),
            ],
            [helper.make_tensor_value_info("Z", onnx_dtype, None)],
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [
            torch.randn(*input_shapes[0], dtype=dtype, device=DEVICE),
            torch.randn(*input_shapes[1], dtype=dtype, device=DEVICE),
        ]

    _run_op_test(op[0], onnx_dtype, _create_model, _gen_inputs, **op[1])


@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shapes", [([1024, 2], [1024, 2]), ([2, 3, 3, 3], [3, 1, 3], [2, 1, 3, 1])])
def test_sum_op(onnx_dtype, input_shapes):
    def _create_model(op_type, onnx_dtype):
        graph = helper.make_graph(
            [helper.make_node(op_type, [f"I{idx}" for idx in range(len(input_shapes))], ["O"], name="test")],
            "test",
            [helper.make_tensor_value_info(f"I{idx}", onnx_dtype, None) for idx in range(len(input_shapes))],
            [helper.make_tensor_value_info("O", onnx_dtype, None)],
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [torch.randn(*input_shape, dtype=dtype, device=DEVICE) for input_shape in input_shapes]

    _run_op_test("Sum", onnx_dtype, _create_model, _gen_inputs)


@pytest.mark.parametrize(
    "op",
    [
        ("Sqrt", {}),
        ("Exp", {}),
        ("com.microsoft::Gelu", {}),
        ("com.microsoft::QuickGelu", {}),
        ("com.microsoft::QuickGelu", {"alpha": 1.0}),
    ],
)
@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shape", [[1024, 4], [2, 3, 3, 3], [2049, 1]])
def test_unary_elementwise_op(op, onnx_dtype, input_shape):
    def _create_model(op_type, onnx_dtype, **kwargs):
        domain = ""
        if "::" in op_type:
            domain, op_type = op_type.split("::")
        graph = helper.make_graph(
            [helper.make_node(op_type, ["X"], ["Y"], name="test", domain=domain, **kwargs)],
            "test",
            [helper.make_tensor_value_info("X", onnx_dtype, None)],
            [helper.make_tensor_value_info("Y", onnx_dtype, None)],
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [torch.rand(*input_shape, dtype=dtype, device=DEVICE)]

    _run_op_test(op[0], onnx_dtype, _create_model, _gen_inputs, **op[1])


@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shape_and_exponent", [([1024, 2], 2.0), ([2, 3, 3, 3], 0.5), ([2049], 3.0)])
def test_pow_op(onnx_dtype, input_shape_and_exponent):
    def _create_model(op_type, onnx_dtype):
        graph = helper.make_graph(
            [helper.make_node(op_type, ["X", "Y"], ["Z"], name="test")],
            "test",
            [
                helper.make_tensor_value_info("X", onnx_dtype, None),
                helper.make_tensor_value_info("Y", onnx_dtype, None),
            ],
            [helper.make_tensor_value_info("Z", onnx_dtype, None)],
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [
            torch.rand(*input_shape_and_exponent[0], dtype=dtype, device=DEVICE),
            torch.tensor(input_shape_and_exponent[1], dtype=dtype, device=DEVICE),
        ]

    _run_op_test("Pow", onnx_dtype, _create_model, _gen_inputs)


@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shape", [[1024, 2], [2, 3, 3, 3], [1, 2050]])
def test_cast_op(onnx_dtype, input_shape):
    def _create_model(op_type, onnx_dtype, **kwargs):
        graph = helper.make_graph(
            [helper.make_node(op_type, ["X"], ["Y"], name="test", **kwargs)],
            "test",
            [
                helper.make_tensor_value_info("X", onnx_dtype, None),
            ],
            [helper.make_tensor_value_info("Y", kwargs.get("to"), None)],
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [torch.randn(*input_shape, dtype=dtype, device=DEVICE)]

    kwargs = {"to": TensorProto.FLOAT16 if onnx_dtype == TensorProto.FLOAT else TensorProto.FLOAT}
    _run_op_test("Cast", onnx_dtype, _create_model, _gen_inputs, **kwargs)


@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shapes", [([2, 1024], [2, 1024], [2, 1024]), ([2, 1, 3, 1], [2, 3, 3, 3], [3, 1, 3])])
def test_where_op(onnx_dtype, input_shapes):
    def _create_model(op_type, onnx_dtype):
        graph = helper.make_graph(
            [helper.make_node(op_type, ["C", "X", "Y"], ["Z"], name="test")],
            "test",
            [
                helper.make_tensor_value_info("C", TensorProto.BOOL, None),
                helper.make_tensor_value_info("X", onnx_dtype, None),
                helper.make_tensor_value_info("Y", onnx_dtype, None),
            ],
            [helper.make_tensor_value_info("Z", onnx_dtype, None)],
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [
            torch.rand(*input_shapes[0], dtype=dtype, device=DEVICE) < 0.5,
            torch.randn(*input_shapes[1], dtype=dtype, device=DEVICE),
            torch.randn(*input_shapes[2], dtype=dtype, device=DEVICE),
        ]

    _run_op_test("Where", onnx_dtype, _create_model, _gen_inputs)


@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shape_and_ratio", [([1024, 2], 0.2), ([25, 75], 0.6), ([1, 2049], 0.5)])
def test_dropout_op(onnx_dtype, input_shape_and_ratio):
    graph = helper.make_graph(
        [
            helper.make_node("Add", ["X1", "X2"], ["T1"], name="add1"),
            helper.make_node("Dropout", ["T1", "ratio"], ["Y1", "mask1"], name="dropout1"),
            helper.make_node("Add", ["Y1", "X3"], ["T2"], name="add2"),
            helper.make_node("Dropout", ["T2", "ratio"], ["Y2", "mask2"], name="dropout2"),
        ],
        "test",
        [
            helper.make_tensor_value_info("X1", onnx_dtype, None),
            helper.make_tensor_value_info("X2", onnx_dtype, None),
            helper.make_tensor_value_info("X3", onnx_dtype, None),
        ],
        [
            helper.make_tensor_value_info("Y1", onnx_dtype, None),
            helper.make_tensor_value_info("mask1", TensorProto.BOOL, None),
            helper.make_tensor_value_info("Y2", onnx_dtype, None),
            helper.make_tensor_value_info("mask2", TensorProto.BOOL, None),
        ],
        initializer=[helper.make_tensor("ratio", TensorProto.FLOAT, (), [input_shape_and_ratio[1]])],
    )
    model_str = helper.make_model(graph, producer_name="test").SerializeToString()
    torch_dtype = _onnx_dtype_to_torch_dtype(onnx_dtype)
    input_tensor = [
        torch.randn(*input_shape_and_ratio[0], dtype=torch_dtype, device=DEVICE),
        torch.randn(*input_shape_and_ratio[0], dtype=torch_dtype, device=DEVICE),
        torch.randn(*input_shape_and_ratio[0], dtype=torch_dtype, device=DEVICE),
    ]
    outputs = call_triton_by_onnx(hash(model_str), model_str, *[to_dlpack(t) for t in input_tensor])
    y1, mask1, y2, mask2 = tuple([_from_dlpack(o).detach().cpu().numpy().flatten() for o in outputs])
    x1 = (input_tensor[0] + input_tensor[1]).detach().cpu().numpy().flatten()
    x2 = y1 * mask1 + input_tensor[2].detach().cpu().numpy().flatten()

    def _check_output(x, y, mask, ratio):
        all_count = 0
        masked_count = 0
        for x_value, y_value, mask_value in zip(x, y, mask, strict=False):
            if mask_value:
                assert abs(y_value - x_value / (1.0 - ratio)) < 0.05
            else:
                assert y_value == 0
            if not mask_value:
                masked_count += 1
            all_count += 1
        assert abs(masked_count / all_count - ratio) < 0.05

    _check_output(x1, y1, mask1, input_shape_and_ratio[1])
    _check_output(x2, y2, mask2, input_shape_and_ratio[1])
    assert any(mask1[i] != mask2[i] for i in range(len(mask1)))


@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shape_and_ratio", [([1024, 2], 0.2), ([25, 75], 0.6), ([1, 2049], 0.5)])
def test_dropout_grad_op(onnx_dtype, input_shape_and_ratio):
    def _create_model(op_type, onnx_dtype, **kwargs):
        graph = helper.make_graph(
            [
                helper.make_node(op_type, ["dY", "mask", "ratio"], ["dX"], name="test", domain="com.microsoft"),
            ],
            "test",
            [
                helper.make_tensor_value_info("dY", onnx_dtype, None),
                helper.make_tensor_value_info("mask", TensorProto.BOOL, None),
            ],
            [helper.make_tensor_value_info("dX", onnx_dtype, None)],
            initializer=[helper.make_tensor("ratio", TensorProto.FLOAT, (), [input_shape_and_ratio[1]])],
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [
            torch.randn(*input_shape_and_ratio[0], dtype=dtype, device=DEVICE),
            torch.rand(*input_shape_and_ratio[0], dtype=dtype, device=DEVICE) >= input_shape_and_ratio[1],
        ]

    kwargs = {"ratio": input_shape_and_ratio[1]}
    _run_op_test("DropoutGrad", onnx_dtype, _create_model, _gen_inputs, **kwargs)


@pytest.mark.parametrize("op_type", ["ReduceSum", "ReduceMean", "ReduceMax", "ReduceMin"])
@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize(
    "input_shape_and_reduce_info",
    [
        ([2, 1024], [-1], True),
        ([1050, 3], [0], False),
        ([2, 3, 3, 3], [1, 2], True),
        ([123, 4, 5, 6], [2], False),
        ([16, 8, 16, 8], [1, 3], True),
        ([16, 8, 16, 8], [0, 2], False),
        ([16, 8, 16, 8], [0, 1, 2, 3], True),
        ([16, 1, 16, 8], [0, 1, 2, 3], False),
    ],
)
def test_reduce_op(op_type, onnx_dtype, input_shape_and_reduce_info):
    def _create_model(op_type, onnx_dtype, **kwargs):
        reduce_inputs = ["X"]
        initializer = []
        if input_shape_and_reduce_info[1] is not None:
            reduce_inputs.append("axes")
            initializer.append(
                helper.make_tensor(
                    "axes",
                    onnx.TensorProto.INT64,
                    [len(input_shape_and_reduce_info[1])],
                    input_shape_and_reduce_info[1],
                )
            )
        node = (
            helper.make_node(op_type, reduce_inputs, ["Y"], name="test", keepdims=input_shape_and_reduce_info[2])
            if input_shape_and_reduce_info[2] is not None
            else helper.make_node(op_type, reduce_inputs, ["Y"], name="test")
        )
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", onnx_dtype, None)],
            [helper.make_tensor_value_info("Y", onnx_dtype, None)],
            initializer=initializer,
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [torch.randn(*input_shape_and_reduce_info[0], dtype=dtype, device=DEVICE)]

    kwargs = {}
    if input_shape_and_reduce_info[1] is not None:
        kwargs["axes"] = input_shape_and_reduce_info[1]
    if input_shape_and_reduce_info[2] is not None:
        kwargs["keepdims"] = input_shape_and_reduce_info[2]
    if onnx_dtype == TensorProto.FLOAT16:
        kwargs["atol"] = 1e-2
        kwargs["rtol"] = 1e-2
    _run_op_test(op_type, onnx_dtype, _create_model, _gen_inputs, **kwargs)


@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shape_and_axis", [([2, 1024], -1), ([2, 3, 3, 3], 1), ([4, 2049], 1)])
def test_softmax_op(onnx_dtype, input_shape_and_axis):
    def _create_model(op_type, onnx_dtype, **kwargs):
        graph = helper.make_graph(
            [helper.make_node(op_type, ["X"], ["Y"], name="test", **kwargs)],
            "test",
            [helper.make_tensor_value_info("X", onnx_dtype, None)],
            [helper.make_tensor_value_info("Y", onnx_dtype, None)],
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [torch.randn(*input_shape_and_axis[0], dtype=dtype, device=DEVICE)]

    kwargs = {"axis": input_shape_and_axis[1]}
    _run_op_test("Softmax", onnx_dtype, _create_model, _gen_inputs, **kwargs)


@pytest.mark.parametrize("onnx_dtype", [TensorProto.FLOAT, TensorProto.FLOAT16])
@pytest.mark.parametrize("input_shape_and_axis", [([2, 1024], -1), ([2, 3, 3, 3], 2), ([4, 2049], 1)])
def test_layer_norm_op(onnx_dtype, input_shape_and_axis):
    def _create_model(op_type, onnx_dtype, **kwargs):
        graph = helper.make_graph(
            [helper.make_node(op_type, ["X", "W", "B"], ["Y"], name="test", **kwargs)],
            "test",
            [
                helper.make_tensor_value_info("X", onnx_dtype, None),
                helper.make_tensor_value_info("W", onnx_dtype, None),
                helper.make_tensor_value_info("B", onnx_dtype, None),
            ],
            [helper.make_tensor_value_info("Y", onnx_dtype, None)],
        )
        return helper.make_model(graph, producer_name="test")

    def _gen_inputs(dtype):
        return [
            torch.randn(*input_shape_and_axis[0], dtype=dtype, device=DEVICE),
            torch.randn(*input_shape_and_axis[0][input_shape_and_axis[1] :], dtype=dtype, device=DEVICE),
            torch.randn(input_shape_and_axis[0][input_shape_and_axis[1] :], dtype=dtype, device=DEVICE),
        ]

    kwargs = {"axis": input_shape_and_axis[1]}
    _run_op_test("LayerNormalization", onnx_dtype, _create_model, _gen_inputs, **kwargs)


@pytest.mark.parametrize("dtype", [torch.float, torch.float16])
@pytest.mark.parametrize(
    "input_info",
    [
        ([32, 64], False, [64, 16], False, 1.0),
        ([33, 68], False, [18, 68], True, 0.5),
        ([128, 64], True, [128, 32], False, 0.5),
        ([123, 234], True, [345, 123], True, -1.0),
        ([22, 33, 44], False, [44, 55], False, 1.0),
        ([22, 33, 44], False, [666, 44], True, 0.2),
        ([22, 33, 44], True, [33, 666], False, -0.2),
        ([64, 128], False, [16, 64, 128], True, 0.5),
        ([16, 32, 64], False, [16, 64, 32], False, 1.0),
        ([8, 16, 32, 16], True, [8, 16, 32, 32], True, 1.0),
    ],
)
def test_matmul(dtype, input_info):
    pt_inputs = [
        torch.rand(*input_info[0], dtype=dtype, device=DEVICE),
        torch.rand(*input_info[2], dtype=dtype, device=DEVICE),
    ]
    ort_inputs = copy.deepcopy(pt_inputs)
    kwargs = {}
    if input_info[1]:
        pt_inputs[0] = pt_inputs[0].transpose(-1, -2)
        kwargs["trans_a"] = True
    if input_info[3]:
        pt_inputs[1] = pt_inputs[1].transpose(-1, -2)
        kwargs["trans_b"] = True
    if input_info[4] != 1.0:
        kwargs["alpha"] = input_info[4]
    pt_output = torch.matmul(*pt_inputs) * input_info[4]
    alloc_out = random.choice([True, False])
    if alloc_out:
        ort_output = torch.empty(pt_output.shape, dtype=dtype, device=DEVICE)
        ort_inputs.append(ort_output)
        call_triton_by_name("triton_matmul_out", *[to_dlpack(tensor) for tensor in ort_inputs], **kwargs)
    else:
        ort_output = _from_dlpack(
            call_triton_by_name("triton_matmul", *[to_dlpack(tensor) for tensor in ort_inputs], **kwargs)
        )
    rtol = 1e-02 if dtype == torch.float16 else 1e-04
    atol = 1e-02 if dtype == torch.float16 else 1e-05
    _test_helpers.assert_values_are_close(pt_output, ort_output, rtol=rtol, atol=atol)


@pytest.mark.parametrize("dtype", [torch.float, torch.float16])
@pytest.mark.parametrize(
    "input_info",
    [
        ([64, 32], False, [32, 64], False, [64, 64], 1.0, 0.0),
        ([65, 129], False, [65, 129], True, [65, 1], 0.5, -0.5),
        ([127, 63], True, [127, 127], False, [127], -1.0, 0.2),
        ([256, 64], True, [128, 256], True, [1], 0.2, 0.5),
    ],
)
def test_gemm(dtype, input_info):
    pt_inputs = [
        torch.rand(*input_info[0], dtype=dtype, device=DEVICE),
        torch.rand(*input_info[2], dtype=dtype, device=DEVICE),
        torch.rand(*input_info[4], dtype=dtype, device=DEVICE),
    ]
    ort_inputs = copy.deepcopy(pt_inputs)
    kwargs = {}
    if input_info[1]:
        pt_inputs[0] = pt_inputs[0].transpose(-1, -2)
        kwargs["trans_a"] = True
    if input_info[3]:
        pt_inputs[1] = pt_inputs[1].transpose(-1, -2)
        kwargs["trans_b"] = True
    if input_info[5] != 1.0:
        kwargs["alpha"] = input_info[5]
    if input_info[6] != 1.0:
        kwargs["beta"] = input_info[6]
    pt_output = torch.matmul(pt_inputs[0], pt_inputs[1]) * input_info[5] + pt_inputs[2] * input_info[6]
    alloc_out = random.choice([True, False])
    if alloc_out:
        ort_output = torch.empty(pt_output.shape, dtype=dtype, device=DEVICE)
        ort_inputs.append(ort_output)
        call_triton_by_name("triton_gemm_out", *[to_dlpack(tensor) for tensor in ort_inputs], **kwargs)
    else:
        ort_output = _from_dlpack(
            call_triton_by_name("triton_gemm", *[to_dlpack(tensor) for tensor in ort_inputs], **kwargs)
        )
    rtol = 1e-02 if dtype == torch.float16 else 1e-04
    atol = 1e-02 if dtype == torch.float16 else 1e-05
    _test_helpers.assert_values_are_close(pt_output, ort_output, rtol=rtol, atol=atol)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float16])
def test_elementwise_module(dtype):
    n, d, h, w = 8, 768, 12, 64

    class NeuralNetElementwise(torch.nn.Module):
        def forward(self, input1, input2, input3, input4):
            return input1 + input2 - input3 * input4

    def _gen_inputs(dtype):
        return [
            torch.rand(n, d, h, w, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(w, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(d, 1, 1, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(n, 1, h, w, dtype=dtype, device=DEVICE, requires_grad=True),
        ]

    _run_module_test(NeuralNetElementwise, dtype, _gen_inputs, 1)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float16])
@pytest.mark.parametrize("input_shapes_and_axis", [([2, 3, 3, 3], [3, 3], 2), ([2, 1024], [2, 1024], -1)])
def test_softmax_module(dtype, input_shapes_and_axis):
    class NeuralNetSoftmax(torch.nn.Module):
        def forward(self, input1, input2):
            return torch.softmax(input1 * input2, dim=input_shapes_and_axis[2])

    def _gen_inputs(dtype):
        return [
            torch.randn(*input_shapes_and_axis[0], dtype=dtype, device=DEVICE, requires_grad=True),
            torch.randn(*input_shapes_and_axis[1], dtype=dtype, device=DEVICE, requires_grad=True),
        ]

    _run_module_test(NeuralNetSoftmax, dtype, _gen_inputs, 2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float16])
@pytest.mark.parametrize(
    "input_shapes_and_axis", [([2, 1024], [2, 1024], -1), ([2, 2049], [2, 1], -1), ([2, 3, 3, 3], [3, 3], 2)]
)
def test_layer_norm_module(dtype, input_shapes_and_axis):
    pytest.skip("LayerNorm is disabled for now due to perf issue.")

    class NeuralNetLayerNorm(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.layer_norm = torch.nn.LayerNorm(
                input_shapes_and_axis[0][input_shapes_and_axis[2] :], device=DEVICE, dtype=dtype
            )

        def forward(self, input1, input2):
            return self.layer_norm(input1 * input2)

    def _gen_inputs(dtype):
        return [
            torch.randn(*input_shapes_and_axis[0], dtype=dtype, device=DEVICE, requires_grad=True),
            torch.randn(*input_shapes_and_axis[1], dtype=dtype, device=DEVICE, requires_grad=True),
        ]

    _run_module_test(NeuralNetLayerNorm, dtype, _gen_inputs, 2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float16])
def test_dynamic_shapes_elementwise_module(dtype):
    class NeuralNetSymbolicShapesElementwise(torch.nn.Module):
        def forward(self, x, y, u, v):
            return x * y - (u + v)

    def _gen_inputs(dtype):
        dim1 = 64 * random.randint(2, 4)
        dim2 = 64 * random.randint(2, 4)
        return [
            torch.rand(16, dim1, dim2, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(16, 1, dim2, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(dim1, 1, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(16, dim1, dim2, dtype=dtype, device=DEVICE, requires_grad=True),
        ]

    _run_module_test(NeuralNetSymbolicShapesElementwise, dtype, _gen_inputs, 1)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float16])
def test_dynamic_shapes_reduction_module(dtype):
    class NeuralNetSymbolicShapesReduction(torch.nn.Module):
        def forward(self, x, y, z):
            return torch.softmax(x * y + z, dim=-1)

    def _gen_inputs(dtype):
        dim1 = 64 * random.randint(2, 4)
        dim2 = 64 * random.randint(2, 4)
        return [
            torch.rand(16, dim1, dim2, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(16, 1, dim2, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(dim1, 1, dtype=dtype, device=DEVICE, requires_grad=True),
        ]

    _run_module_test(NeuralNetSymbolicShapesReduction, dtype, _gen_inputs, 2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float16])
@pytest.mark.parametrize("has_sum", [True, False])
def test_slice_scel_module(dtype, has_sum):
    class NeuralNetSliceScel(torch.nn.Module):
        def forward(self, logits, labels):
            shift_logits = logits[..., :-1, :].contiguous()
            labels = labels[..., 1:].contiguous()
            loss_fct = torch.nn.CrossEntropyLoss()
            loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), labels.view(-1))
            return logits + loss if has_sum else loss

    def _gen_inputs(dtype):
        return [
            (torch.rand(4, 8, 16) * 0.01).to(dtype=dtype, device=DEVICE).requires_grad_(True),
            torch.randint(0, 16, (4, 8), dtype=torch.int64, device=DEVICE),
        ]

    _run_module_test(NeuralNetSliceScel, dtype, _gen_inputs, 2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float16])
@pytest.mark.parametrize("input_shapes", [([128, 64], [64, 64]), ([16, 64, 128], [16, 128, 64])])
def test_matmul_tunable_op(dtype, input_shapes):
    class NeuralNetMatmul(torch.nn.Module):
        def forward(self, input1, input2):
            return torch.matmul(input1, input2)

    def _gen_inputs(dtype):
        return [
            torch.rand(*input_shapes[0], dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(*input_shapes[1], dtype=dtype, device=DEVICE, requires_grad=True),
        ]

    _run_tunable_op_test(NeuralNetMatmul, dtype, _gen_inputs, "MatMulTunableOp", 2)


@pytest.mark.parametrize("dtype", [torch.float32, torch.float16])
@pytest.mark.parametrize("m_n_k", [(64, 64, 64)])
def test_gemm_tunable_op(dtype, m_n_k):
    class NeuralNetGemm(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.linear = torch.nn.Linear(m_n_k[2], m_n_k[1])

        def forward(self, input):
            return self.linear(input)

    def _gen_inputs(dtype):
        return [torch.rand(m_n_k[0], m_n_k[2], dtype=dtype, device=DEVICE, requires_grad=True)]

    _run_tunable_op_test(NeuralNetGemm, dtype, _gen_inputs, "GemmTunableOp", 2)


def test_user_config():
    n, d, h, w = 8, 768, 12, 64
    dtype = torch.float32

    class NeuralNetElementwise(torch.nn.Module):
        def forward(self, input1, input2, input3, input4):
            return input1 + input2 - input3 * input4

    def _gen_inputs(dtype):
        return [
            torch.rand(n, d, h, w, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(w, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(d, 1, 1, dtype=dtype, device=DEVICE, requires_grad=True),
            torch.rand(n, 1, h, w, dtype=dtype, device=DEVICE, requires_grad=True),
        ]

    user_config = (
        '{"ops": {"Add": {"versions": [13, 14]}, "Mul": {"versions": [13, 14]}}, '
        '"initializer": "scalar", "min_nodes": 2}'
    )
    with open("user_config.json", "w", encoding="UTF-8") as f:
        f.write(user_config)
    os.environ["ORTMODULE_TRITON_CONFIG_FILE"] = "./user_config.json"

    # Mul is not supported, the graph is splited to 2 subgraphs with single Op, which will not be fused to TritonOp.
    _run_module_test(NeuralNetElementwise, dtype, _gen_inputs, 0)

    del os.environ["ORTMODULE_TRITON_CONFIG_FILE"]
    os.remove(os.path.join(os.getcwd(), "user_config.json"))



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import logging
import sys

from _test_commons import run_subprocess

logging.basicConfig(format="%(asctime)s %(name)s [%(levelname)s] - %(message)s", level=logging.DEBUG)
log = logging.getLogger("ORTModuleDistributedTests")


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--cwd", help="Path to the current working directory")
    parser.add_argument("--mnist", help="Path to the mnist data directory", type=str, default=None)
    return parser.parse_args()


def run_ortmodule_deepspeed_zero_stage_1_tests(cwd, log, data_dir):
    log.debug("Running: ORTModule deepspeed zero stage 1 tests")

    command = [
        "deepspeed",
        "orttraining_test_ortmodule_deepspeed_zero_stage_1.py",
        "--deepspeed_config",
        "orttraining_test_ortmodule_deepspeed_zero_stage_1_config.json",
    ]

    if data_dir:
        command.extend(["--data-dir", data_dir])

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_pytorch_ddp_tests(cwd, log):
    log.debug("Running: ORTModule Pytorch DDP tests")

    command = [sys.executable, "orttraining_test_ortmodule_pytorch_ddp.py", "--use_ort_module"]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_ortmodule_deepspeed_pipeline_parallel_tests(cwd, log):
    log.debug("Running: ORTModule deepspeed pipeline parallel tests")

    command = [
        "deepspeed",
        "orttraining_test_ortmodule_deepspeed_pipeline_parallel.py",
        "--deepspeed_config",
        "orttraining_test_ortmodule_deepspeed_pipeline_parallel_config.json",
    ]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_ort_pipeline_module_tests(cwd, log, layer_spec_flag=False):
    log.debug("Running: ORTPipelineModule tests")

    command = [
        "deepspeed",
        "orttraining_test_ort_pipeline_module.py",
        "--deepspeed_config",
        "orttraining_test_ortmodule_deepspeed_pipeline_parallel_config.json",
    ]

    if layer_spec_flag:
        command.append("--layer_spec=True")

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_ortmodule_fairscale_sharded_optimizer_tests(cwd, log, data_dir):
    log.debug("Running: ORTModule fairscale sharded optimizer tests")
    command = [
        "python3",
        "orttraining_test_ortmodule_fairscale_sharded_optimizer.py",
        "--use_sharded_optimizer",
        "--use_ortmodule",
    ]
    if data_dir:
        command.extend(["--data-dir", data_dir])

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def run_distributed_cache_test(cwd, log):
    log.debug("Running: ORTModule Cache Test")

    command = [
        "torchrun",
        "--nproc_per_node",
        "2",
        "orttraining_test_ortmodule_cache.py",
    ]

    run_subprocess(command, cwd=cwd, log=log).check_returncode()


def main():
    args = parse_arguments()
    cwd = args.cwd

    log.info("Running ortmodule tests pipeline")

    run_pytorch_ddp_tests(cwd, log)

    run_ortmodule_deepspeed_zero_stage_1_tests(cwd, log, args.mnist)

    run_ortmodule_deepspeed_pipeline_parallel_tests(cwd, log)

    # Deepspeed ORTPipelineModule Tests
    run_ort_pipeline_module_tests(cwd, log)
    run_ort_pipeline_module_tests(cwd, log, layer_spec_flag=True)

    run_ortmodule_fairscale_sharded_optimizer_tests(cwd, log, args.mnist)

    run_distributed_cache_test(cwd, log)
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


import copy
import os
import sys

import _test_helpers
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from torch.nn.parallel import DistributedDataParallel as DDP  # noqa: N817
from torch.nn.parameter import Parameter

import onnxruntime
from onnxruntime.training.ortmodule import ORTModule
from onnxruntime.training.ortmodule._graph_execution_manager_factory import GraphExecutionManagerFactory  # noqa: F401

torch.manual_seed(1)
onnxruntime.set_seed(1)


class ReduceWithMarkDirtyFunction(torch.autograd.Function):
    # All-reduce the input from the model parallel region.
    @staticmethod
    def forward(ctx, arg):
        def reduce(buffer):
            # All-reduce.
            address_for_torch_tensor = int(id(buffer))
            torch.distributed.all_reduce(buffer)
            address_for_output_torch_tensor = int(id(buffer))
            if address_for_output_torch_tensor != address_for_torch_tensor:
                raise ValueError("The output torch tensor should reuse the input torch tensor, but actually not.")
            return buffer

        ctx.save_for_backward(arg)
        ctx.mark_dirty(arg)
        return reduce(arg)

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output.add(1.0)


class ReduceWithMarkDirtyModel(torch.nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.reduce_op_ = ReduceWithMarkDirtyFunction.apply
        self.bias = Parameter(torch.empty(dim, device=torch.cuda.current_device(), dtype=torch.float))

        # Always initialize bias to zero.
        with torch.no_grad():
            self.bias.uniform_()

    def forward(self, model_input):
        x = model_input + self.bias
        y1 = self.reduce_op_(x)  # at this point x require_grad = True
        y2 = x.add(self.bias)
        out = x + y1 + y2
        return out


def test_Distributed_ReduceWithMarkDirtyModel(rank, size):
    def run_with_pytorch_on_gpu(model, args, rank, device):
        model.to(device)
        cuda_args = [input_.to(device) for input_ in args]
        model = DDP(model, device_ids=[rank])
        output = model(*cuda_args)
        output.sum().backward()
        return output, [arg.grad for arg in cuda_args]

    def run_with_ort_on_gpu(model, args, rank, device):
        model.to(device)
        model = ORTModule(model)

        _test_helpers.set_onnx_fallthrough_export_type(model)
        model = DDP(model, device_ids=[rank])
        cuda_args = [arg.to(device) for arg in args]
        output = model(*cuda_args)
        output.sum().backward()
        return output, [arg.grad for arg in cuda_args]

    try:
        torch.cuda.set_device("cuda:" + str(rank))
        os.environ["MASTER_ADDR"] = "127.0.0.1"
        os.environ["MASTER_PORT"] = "29500"
        dist.init_process_group(
            backend="nccl", init_method="tcp://" + os.environ["MASTER_ADDR"] + ":23456", world_size=size, rank=rank
        )

        dim = 32
        device = torch.device("cuda:" + str(rank))
        x = torch.randn(dim, dtype=torch.float)
        x.requires_grad = True
        x_copy = copy.deepcopy(x)
        m = ReduceWithMarkDirtyModel(dim)

        torch.cuda.synchronize()

        outputs, grads = run_with_pytorch_on_gpu(m, [x], rank, device)

        torch.cuda.synchronize()

        outputs_ort, grads_ort = run_with_ort_on_gpu(m, [x_copy], rank, device)

        torch.cuda.synchronize()

        val_list_a = [o.detach().cpu() for o in outputs if o is not None]
        val_list_b = [o.detach().cpu() for o in outputs_ort if o is not None]
        _test_helpers.compare_tensor_list(val_list_a, val_list_b)

        val_list_a = [o.detach().cpu() for o in grads if o is not None]
        val_list_b = [o.detach().cpu() for o in grads_ort if o is not None]
        _test_helpers.compare_tensor_list(val_list_a, val_list_b)
    except Exception as e:
        print(
            f"test_Distributed_ReduceWithMarkDirtyModel fail with rank {rank} with world size {size} with exception: \n{e}."
        )
        raise e


if __name__ == "__main__":
    size = 2
    try:
        mp.spawn(test_Distributed_ReduceWithMarkDirtyModel, nprocs=size, args=(size,))
    except Exception:
        sys.stdout.flush()
        sys.stderr.flush()
        raise



import argparse
import glob
import os
import sys

import numpy as np
import onnx
from onnx import numpy_helper


def read_tensorproto_pb_file(filename):
    """Return tuple of tensor name and numpy.ndarray of the data from a pb file containing a TensorProto."""

    tensor = onnx.load_tensor(filename)
    np_array = numpy_helper.to_array(tensor)
    return tensor.name, np_array


def dump_tensorproto_pb_file(filename):
    """Dump the data from a pb file containing a TensorProto."""

    name, data = read_tensorproto_pb_file(filename)
    print(f"Name: {name}")
    print(f"Shape: {data.shape}")
    print(data)


def dump_pb(dir_or_filename):
    """Dump the data from either a single .pb file, or all .pb files in a directory.
    All files must contain a serialized TensorProto."""

    if os.path.isdir(dir_or_filename):
        for f in glob.glob(os.path.join(dir_or_filename, "*.pb")):
            print(f)
            dump_tensorproto_pb_file(f)
    else:
        dump_tensorproto_pb_file(dir_or_filename)


def numpy_to_pb(name, np_data, out_filename):
    """Convert numpy data to a protobuf file."""

    tensor = numpy_helper.from_array(np_data, name)
    onnx.save_tensor(tensor, out_filename)


def image_to_numpy(filename, shape, channels_last, add_batch_dim):
    """Convert an image file into a numpy array."""

    import PIL.Image  # from 'Pillow' package

    img = PIL.Image.open(filename)
    if shape:
        w, h = img.size
        new_w = shape[1]
        new_h = shape[0]

        # use the dimension that needs to shrink the least to resize to an image where that dimension matches the
        # target size.
        w_ratio = new_w / w
        h_ratio = new_h / h
        ratio = max(h_ratio, w_ratio)
        interim_w = int(w * ratio)
        interim_h = int(h * ratio)
        img = img.resize((interim_w, interim_h), PIL.Image.ANTIALIAS)

        # center crop to the final target size
        left = (interim_w - new_w) / 2
        top = (interim_h - new_h) / 2
        right = (interim_w + new_w) / 2
        bottom = (interim_h + new_h) / 2
        img = img.crop((left, top, right, bottom))

    img_as_np = np.array(img).astype(np.float32)
    if not channels_last:
        # HWC to CHW
        img_as_np = np.transpose(img_as_np, (2, 0, 1))

    if add_batch_dim:
        # to NCHW or NHWC
        img_as_np = np.expand_dims(img_as_np, axis=0)

    return img_as_np


def create_random_data(shape, type, minvalue, maxvalue, seed):
    nptype = np.dtype(type)
    np.random.seed(seed)
    return ((maxvalue - minvalue) * np.random.sample(shape) + minvalue).astype(nptype)


def update_name_in_pb(filename, name, output_filename):
    """Update the name of the tensor in the pb file."""

    tensor = onnx.load_tensor(filename)
    tensor.name = name

    if not output_filename:
        output_filename = filename

    onnx.save_tensor(tensor, output_filename)


def get_arg_parser():
    parser = argparse.ArgumentParser(
        description="""
        Utilities for working with the input/output protobuf files used by the ONNX test cases and onnx_test_runner.
        These are expected to only contain a serialized TensorProto.

        dump_pb: Dumps the TensorProto data from an individual pb file, or all pb files in a directory.
        numpy_to_pb: Convert numpy array saved to a file with numpy.save() to a TensorProto, and serialize to a pb file.
        image_to_pb: Convert data from an image file into a TensorProto, and serialize to a pb file.
        random_to_pb: Create a TensorProto with random data, and serialize to a pb file.
        raw_to_pb: Create a uint8 TensorProto with raw data from a file, and serialize to a pb file.
        string_to_pb: Create a string TensorProto with the input string, and serialize to a pb file.
        update_name_in_pb: Update the TensorProto.name value in a pb file.
                           Updates the input file unless --output <filename> is specified.
        """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--action",
        help="Action to perform",
        choices=[
            "dump_pb",
            "numpy_to_pb",
            "image_to_pb",
            "random_to_pb",
            "raw_to_pb",
            "string_to_pb",
            "update_name_in_pb",
        ],
        required=True,
    )

    parser.add_argument("--input", help="The input filename, directory name or string.")
    parser.add_argument("--name", help="The value to set TensorProto.name to if creating/updating one.")
    parser.add_argument("--output", help="Filename to serialize the TensorProto to.")

    image_to_pb_group = parser.add_argument_group("image_to_pb", "image_to_pb specific options")
    image_to_pb_group.add_argument(
        "--raw",
        action="store_true",
        help="Save raw image bytes for usage with a model that has the DecodeImage custom operator "
        "from onnxruntime-extensions.",
    )
    image_to_pb_group.add_argument(
        "--resize",
        default=None,
        type=lambda s: [int(item) for item in s.split(",")],
        help="Provide the height and width to resize to as comma separated values."
        " e.g. --shape 200,300 will resize to height 200 and width 300.",
    )
    image_to_pb_group.add_argument(
        "--channels_last", action="store_true", help="Transpose image from channels first to channels last."
    )
    image_to_pb_group.add_argument(
        "--add_batch_dim",
        action="store_true",
        help="Prepend a batch dimension with value of 1 to the shape. i.e. convert from CHW to NCHW",
    )

    random_to_pb_group = parser.add_argument_group("random_to_pb", "random_to_pb specific options")
    random_to_pb_group.add_argument(
        "--shape",
        type=lambda s: [int(item) for item in s.split(",")],
        help="Provide the shape as comma separated values e.g. --shape 200,200",
    )
    random_to_pb_group.add_argument(
        "--datatype",
        help="numpy dtype value for the data type. e.g. f4=float32, i8=int64. "
        "See: https://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html",
    )
    random_to_pb_group.add_argument(
        "--min_value", default=0, type=int, help="Limit the generated values to this minimum."
    )
    random_to_pb_group.add_argument(
        "--max_value", default=1, type=int, help="Limit the generated values to this maximum."
    )
    random_to_pb_group.add_argument(
        "--seed", default=None, type=int, help="seed to use for the random values so they're deterministic."
    )

    return parser


if __name__ == "__main__":
    arg_parser = get_arg_parser()
    args = arg_parser.parse_args()

    if args.action == "dump_pb":
        if not args.input:
            print("Missing argument. Need input to be specified.", file=sys.stderr)
            sys.exit(-1)
        np.set_printoptions(precision=10)
        dump_pb(args.input)
    elif args.action == "numpy_to_pb":
        if not args.input or not args.output or not args.name:
            print("Missing argument. Need input, output and name to be specified.", file=sys.stderr)
            sys.exit(-1)
        # read data saved with numpy
        data = np.load(args.input)
        numpy_to_pb(args.name, data, args.output)
    elif args.action == "image_to_pb":
        if not args.input or not args.output or not args.name:
            print("Missing argument. Need input, output, name to be specified.", file=sys.stderr)
            sys.exit(-1)

        if args.raw:
            img_np = np.fromfile(args.input, np.ubyte)
        else:
            img_np = image_to_numpy(args.input, args.resize, args.channels_last, args.add_batch_dim)

        numpy_to_pb(args.name, img_np, args.output)
    elif args.action == "random_to_pb":
        if not args.output or not args.shape or not args.datatype or not args.name:
            print("Missing argument. Need output, shape, datatype and name to be specified.", file=sys.stderr)
            sys.exit(-1)

        data = create_random_data(args.shape, args.datatype, args.min_value, args.max_value, args.seed)
        numpy_to_pb(args.name, data, args.output)
    elif args.action == "raw_to_pb":
        if not args.input or not args.output or not args.name:
            print("Missing argument. Need input, output and name to be specified.", file=sys.stderr)
            sys.exit(-1)

        data = np.fromfile(args.input, dtype=np.ubyte)
        numpy_to_pb(args.name, data, args.output)
    elif args.action == "string_to_pb":
        if not args.input or not args.output or not args.name:
            print("Missing argument. Need input, output and name to be specified.", file=sys.stderr)
            sys.exit(-1)

        data = np.ndarray((1), dtype=object)
        data[0] = args.input
        numpy_to_pb(args.name, data, args.output)
    elif args.action == "update_name_in_pb":
        if not args.input or not args.name:
            print("Missing argument. Need input and name to be specified.", file=sys.stderr)
            sys.exit(-1)

        update_name_in_pb(args.input, args.name, args.output)
    else:
        print("Unknown action.", file=sys.stderr)
        arg_parser.print_help(sys.stderr)



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import json
import pathlib
import subprocess
import sys

REPO_ROOT = pathlib.Path(__file__).resolve().parents[5]


def parse_args():
    parser = argparse.ArgumentParser(description="Builds ORT and checks the binary size.")

    parser.add_argument("build_check_binsize_config", type=pathlib.Path, help="Path to configuration file.")
    parser.add_argument("--build_dir", type=pathlib.Path, required=True, help="Path to build directory.")
    parser.add_argument("--threshold_size_in_bytes", type=int, help="Binary size limit in bytes.")
    parser.add_argument(
        "--with_debug_info", action="store_true", help="Whether to include debug information in the build."
    )

    return parser.parse_args()


def main():
    args = parse_args()

    with open(args.build_check_binsize_config) as config_file:
        config = json.load(config_file)

    config_type = config["type"]
    os = config["os"]
    arch = config["arch"]
    build_params = config["build_params"]
    build_config = "MinSizeRel"  # could make this configurable if needed
    build_command = (
        [sys.executable, str(REPO_ROOT / "tools/ci_build/build.py"), *build_params]
        + (["--cmake_extra_defines", "ADD_DEBUG_INFO_TO_MINIMAL_BUILD=ON"] if args.with_debug_info else [])
        # put the following options last so they don't get overridden by build_params
        + [f"--build_dir={args.build_dir}", f"--config={build_config}", "--update", "--build", "--parallel", "--test"]
    )

    subprocess.run(build_command, check=True)

    # check binary size
    check_binary_size_command = (
        [
            sys.executable,
            str(REPO_ROOT / "tools/ci_build/github/linux/ort_minimal/check_build_binary_size.py"),
            f"--os={os}",
            f"--arch={arch}",
            f"--build_config={config_type}",
        ]
        + ([f"--threshold={args.threshold_size_in_bytes}"] if args.threshold_size_in_bytes else [])
        + [str(args.build_dir / build_config / "libonnxruntime.so")]
    )

    subprocess.run(check_binary_size_command, check=True)


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
import argparse
import datetime
import os
import sys

import pandas as pd
from azure.kusto.data import KustoConnectionStringBuilder
from azure.kusto.data.data_format import DataFormat
from azure.kusto.ingest import IngestionProperties, QueuedIngestClient, ReportLevel
from perf_utils import (
    avg_ending,
    cpu,
    cuda,
    cuda_fp16,
    fail_name,
    group_title,
    latency_name,
    latency_over_time_name,
    memory_ending,
    memory_name,
    memory_over_time_name,
    model_title,
    op_metrics_name,
    ort_provider_list,
    provider_list,
    second,
    session_name,
    session_over_time_name,
    specs_name,
    standalone_trt,
    standalone_trt_fp16,
    status_name,
    status_over_time_name,
    table_headers,
    trt,
    trt_fp16,
)


def parse_arguments():
    """
    Parses command-line arguments and returns an object with each argument as a field.

    :return: An object whose fields represent the parsed command-line arguments.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("-r", "--report_folder", help="Path to the local file report", required=True)
    parser.add_argument("-c", "--commit_hash", help="Commit hash", required=True)
    parser.add_argument("-u", "--report_url", help="Report Url", required=True)
    parser.add_argument("-t", "--trt_version", help="Tensorrt Version", required=True)
    parser.add_argument("-b", "--branch", help="Branch", required=True)
    parser.add_argument("--kusto_conn", help="Kusto connection URL", required=True)
    parser.add_argument("--database", help="Database name", required=True)
    parser.add_argument("--use_tensorrt_oss_parser", help="Use TensorRT OSS parser", required=False)
    parser.add_argument(
        "-d",
        "--commit_datetime",
        help="Commit datetime in Python's datetime ISO 8601 format",
        required=True,
        type=datetime.datetime.fromisoformat,
    )

    return parser.parse_args()


def adjust_columns(table, columns, db_columns, model_group):
    """
    Utility function that replaces column names in an in-memory table with the appropriate database column names.
    Additionly, this function adds a model group column to all rows in the table.

    :param table: The Pandas table to adjust.
    :param columns: A list of existing column names to rename.
    :param db_columns: A list of databse columns names to use.
    :param model_group: The model group to append as a column.

    :return: The updated table.
    """

    table = table[columns]
    table = table.set_axis(db_columns, axis=1)
    table = table.assign(Group=model_group)
    return table


def get_latency_over_time(report_url, latency_table):
    """
    Returns a new Pandas table with data that tracks the latency of model/EP inference runs over time.

    :param report_url: The URL of the Azure pipeline run/report which produced this latency data.
    :param latency_table: The Pandas table containing per model/EP latencies with the schema:
                          | Model    | ORT-CPUFp32 | ORT-CUDAFp32 | ... |       Group     | ...
                          =====================================================================
                          | resnet.. |    43.61    |     4.18     | ... | onnx-zoo-models | ...

    :return: A new table in which the EPs are not hardcoded as columns. Ex:
             | Model    |      Group      |      Ep      | Latency | ...
             ===========================================================
             | resnet.. | onnx-zoo-models | ORT-CPUFp32  |  43.61  | ...
             | resnet.. | onnx-zoo-models | ORT-CUDAFp32 |  4.18   | ...
    """

    over_time = latency_table.melt(id_vars=[model_title, group_title], var_name="Ep", value_name="Latency")
    over_time = over_time.assign(ReportUrl=report_url)
    over_time = over_time[
        [
            model_title,
            group_title,
            "Ep",
            "Latency",
            "ReportUrl",
        ]
    ]
    over_time.fillna("", inplace=True)
    return over_time


def get_failures(fail, model_group):
    """
    Returns a new Pandas table with data that tracks failed model/EP inference runs.

    :param fail: The Pandas table containing raw failure data imported from a CSV file.
    :param model_group: The model group namespace to append as a column.

    :return: The updated table.
    """

    fail_columns = fail.keys()
    fail_db_columns = [model_title, "Ep", "ErrorType", "ErrorMessage"]
    fail = adjust_columns(fail, fail_columns, fail_db_columns, model_group)
    return fail


def get_memory(memory, model_group):
    """
    Returns a new Pandas table with data that tracks peak memory usage per model/EP.

    :param memory: The Pandas table containing raw memory usage data imported from a CSV file.
    :param model_group: The model group namespace to append as a column.

    :return: The updated table.
    """

    memory_columns = [model_title]
    for provider in provider_list:
        if cpu not in provider:
            memory_columns.append(provider + memory_ending)
    memory_db_columns = [
        model_title,
        cuda,
        trt,
        standalone_trt,
        cuda_fp16,
        trt_fp16,
        standalone_trt_fp16,
    ]
    memory = adjust_columns(memory, memory_columns, memory_db_columns, model_group)
    return memory


def get_memory_over_time(memory_table):
    """
    Returns a new Pandas table with data that tracks the peak memory usage of model/EP inference runs over time.

    :param memory_table: The Pandas table containing per model/EP memory usage with the schema:
                          | Model    | ORT-CUDAFp16 | ORT-CUDAFp32 | ... |       Group     | ...
                          ======================================================================
                          | resnet.. |     685      |     873      | ... | onnx-zoo-models | ...

    :return: A new table in which the EPs are not hardcoded as columns. Ex:
             | Model    |      Group      |      Ep      | MemUsage | ...
             ============================================================
             | resnet.. | onnx-zoo-models | ORT-CUDAFp16 |   685    | ...
             | resnet.. | onnx-zoo-models | ORT-CUDAFp32 |   873    | ...
    """

    over_time = memory_table.melt(id_vars=[model_title, group_title], var_name="Ep", value_name="MemUsage")
    over_time = over_time[
        [
            model_title,
            group_title,
            "Ep",
            "MemUsage",
        ]
    ]

    over_time.fillna("", inplace=True)

    return over_time


def get_session_over_time(session_table):
    """
    Returns a new Pandas table with data that tracks the session creation times of model/EP combinations over time.

    :param session_table: The Pandas table containing per model/EP session creation times with the schema:
                          | Model    | ORT-CUDAFp16 | ... | ORT-CUDAFp16_second |       Group     | ...
                          =============================================================================
                          | resnet.. |     1.99     | ... |         0.92        | onnx-zoo-models | ...

    :return: A new table in which the EPs are not hardcoded as columns. Ex:
             | Model    |      Group      |      Ep      | SessionCreationTime | SessionCreationTime_second | ...
             ====================================================================================================
             | resnet.. | onnx-zoo-models | ORT-CUDAFp16 |        1.99         |             0.92           | ...
    """

    ep_names = [cpu, cuda_fp16, cuda, trt_fp16, trt]
    over_time_1 = session_table.melt(
        id_vars=[model_title, group_title], value_vars=ep_names, var_name="Ep", value_name="SessionCreationTime"
    )
    over_time_2 = session_table.melt(
        id_vars=[model_title, group_title],
        value_vars=[ep + "_second" for ep in ep_names],
        value_name="SessionCreationTime_second",
    )
    over_time = over_time_1.merge(over_time_2[["SessionCreationTime_second"]], left_index=True, right_index=True)
    over_time = over_time[
        [
            model_title,
            group_title,
            "Ep",
            "SessionCreationTime",
            "SessionCreationTime_second",
        ]
    ]

    over_time.fillna("", inplace=True)

    return over_time


def get_status_over_time(status_table):
    """
    Returns a new Pandas table with data that tracks the compatibility of model/EP combinations over time.

    :param status_table: The Pandas table containing per model/EP compatibility ('Pass' or 'Fail') data with the schema:
                          |     Model     | ORT-CUDAFp16 | ORT-CUDAFp32 | ... |       Group     | ...
                          ===========================================================================
                          | FasterRCNN-10 |     Fail     |     Pass     | ... | onnx-zoo-models | ...

    :return: A new table in which the EPs are not hardcoded as columns. Ex:
             |      Model    |      Group      |      Ep      | Pass | ...
             =============================================================
             | FasterRCNN-10 | onnx-zoo-models | ORT-CUDAFp16 |   0  | ...
             | FasterRCNN-10 | onnx-zoo-models | ORT-CUDAFp32 |   1  | ...
    """

    over_time = status_table.melt(id_vars=[model_title, group_title], var_name="Ep", value_name="Pass")
    over_time["Pass"] = over_time["Pass"].map(lambda s: 1 if s == "Pass" else 0)
    over_time = over_time[
        [
            model_title,
            group_title,
            "Ep",
            "Pass",
        ]
    ]

    return over_time


def get_latency(latency, model_group):
    """
    Returns a new Pandas table with data that tracks inference run latency per model/EP.

    :param latency: The Pandas table containing raw latency data imported from a CSV file.
    :param model_group: The model group namespace to append as a column.

    :return: The updated table.
    """

    latency_columns = [model_title]
    for provider in provider_list:
        latency_columns.append(provider + avg_ending)
    latency_db_columns = table_headers
    latency = adjust_columns(latency, latency_columns, latency_db_columns, model_group)
    return latency


def get_status(status, model_group):
    """
    Returns a new Pandas table with data that tracks whether an EP can successfully run a particular model.

    :param status: The Pandas table containing raw model/EP status data imported from a CSV file.
    :param model_group: The model group namespace to append as a column.

    :return: The updated table.
    """

    status_columns = status.keys()
    status_db_columns = table_headers
    status = adjust_columns(status, status_columns, status_db_columns, model_group)
    return status


def get_specs(specs, branch, commit_hash, commit_datetime):
    """
    Returns a new Pandas table with data that tracks the configuration/specs/versions of the hardware and software
    used to gather benchmarking data.

    :param specs: The Pandas table containing raw specs data imported from a CSV file.
    :param branch: The name of the git branch corresponding to the version of ORT used to gather data.
    :param commit_hash: The short git commit hash corresponding to the version of ORT used to gather data.
    :param commit_datetime: The git commit datetime corresponding to the version of ORT used to gather data.

    :return: The updated table.
    """

    init_id = int(specs.tail(1).get(".", 0).iloc[0]) + 1
    specs_additional = pd.DataFrame(
        {
            ".": [init_id, init_id + 1, init_id + 2],
            "Spec": ["Branch", "CommitId", "CommitTime"],
            "Version": [branch, commit_hash, str(commit_datetime)],
        }
    )

    return pd.concat([specs, specs_additional], ignore_index=True)


def get_session(session, model_group):
    """
    Returns a new Pandas table with data that tracks the ORT session creation time for each model/EP combination.

    :param session: The Pandas table containing raw model/EP session timing data imported from a CSV file.
    :param model_group: The model group namespace to append as a column.

    :return: The updated table.
    """

    session_columns = session.keys()
    session_db_columns = [model_title, *ort_provider_list] + [p + second for p in ort_provider_list]
    session = adjust_columns(session, session_columns, session_db_columns, model_group)
    return session


def write_table(
    ingest_client, database_name, table, table_name, upload_time, identifier, branch, commit_id, commit_date
):
    """
    Uploads the provided table to the database. This function also appends the upload time and unique run identifier
    to the table.

    :param ingest_client: An instance of QueuedIngestClient used to initiate data ingestion.
    :param table: The Pandas table to ingest.
    :param table_name: The name of the table in the database.
    :param upload_time: A datetime object denoting the data's upload time.
    :param identifier: An identifier that associates the uploaded data with an ORT commit/date/branch.
    """

    if table.empty:
        return

    # Add upload time and identifier columns to data table.
    table = table.assign(UploadTime=str(upload_time))
    table = table.assign(Identifier=identifier)
    table = table.assign(Branch=branch)
    table = table.assign(CommitId=commit_id)
    table = table.assign(CommitDate=str(commit_date))
    ingestion_props = IngestionProperties(
        database=database_name,
        table=table_name,
        data_format=DataFormat.CSV,
        report_level=ReportLevel.FailuresAndSuccesses,
    )
    # append rows
    ingest_client.ingest_from_dataframe(table, ingestion_properties=ingestion_props)


def get_identifier(commit_datetime, commit_hash, trt_version, branch, use_tensorrt_oss_parser):
    """
    Returns an identifier that associates uploaded data with an ORT commit/date/branch and a TensorRT version.

    :param commit_datetime: The datetime of the ORT commit used to run the benchmarks.
    :param commit_hash: The hash of the ORT commit used to run the benchmarks.
    :param trt_version: The TensorRT version used to run the benchmarks.
    :param branch: The name of the ORT branch used to run the benchmarks.

    :return: A string identifier.
    """

    date = str(commit_datetime.date())  # extract date only
    if use_tensorrt_oss_parser:
        current_dir = os.path.dirname(os.path.abspath(__file__))
        root_dir = os.path.abspath(os.path.join(current_dir, "../../../../.."))
        deps_txt_path = os.path.join(root_dir, "cmake", "deps.txt")
        commit_head = ""
        with open(deps_txt_path) as file:
            for line in file:
                parts = line.split(";")
                if parts[0] == "onnx_tensorrt":
                    url = parts[1]
                    commit = url.split("/")[-1]
                    commit_head = commit[:6]
                    break
        parser = f"oss_{commit_head}"
    else:
        parser = "builtin"
    return "_".join([date, commit_hash, trt_version, parser, branch])


def main():
    """
    Entry point of this script. Uploads data produced by benchmarking scripts to the database.
    """

    args = parse_arguments()

    # connect to database
    kcsb_ingest = KustoConnectionStringBuilder.with_az_cli_authentication(args.kusto_conn)
    ingest_client = QueuedIngestClient(kcsb_ingest)
    identifier = get_identifier(
        args.commit_datetime, args.commit_hash, args.trt_version, args.branch, args.use_tensorrt_oss_parser
    )
    print(f"DB record identifier: {identifier}")
    upload_time = datetime.datetime.now(tz=datetime.timezone.utc).replace(microsecond=0)

    try:
        result_file = args.report_folder

        folders = os.listdir(result_file)
        os.chdir(result_file)

        tables = [
            fail_name,
            memory_name,
            memory_over_time_name,
            latency_name,
            latency_over_time_name,
            status_name,
            status_over_time_name,
            specs_name,
            session_name,
            session_over_time_name,
            op_metrics_name,
        ]

        table_results = {}
        for table_name in tables:
            table_results[table_name] = pd.DataFrame()

        for model_group in folders:
            os.chdir(model_group)
            csv_filenames = os.listdir()
            for csv in csv_filenames:
                table = pd.read_csv(csv)
                if session_name in csv:
                    table_results[session_name] = pd.concat(
                        [table_results[session_name], get_session(table, model_group)], ignore_index=True
                    )
                elif specs_name in csv:
                    table_results[specs_name] = pd.concat(
                        [
                            table_results[specs_name],
                            get_specs(table, args.branch, args.commit_hash, args.commit_datetime),
                        ],
                        ignore_index=True,
                    )
                elif fail_name in csv:
                    table_results[fail_name] = pd.concat(
                        [table_results[fail_name], get_failures(table, model_group)],
                        ignore_index=True,
                    )
                elif latency_name in csv:
                    table_results[memory_name] = pd.concat(
                        [table_results[memory_name], get_memory(table, model_group)],
                        ignore_index=True,
                    )

                    table_results[latency_name] = pd.concat(
                        [table_results[latency_name], get_latency(table, model_group)],
                        ignore_index=True,
                    )
                elif status_name in csv:
                    table_results[status_name] = pd.concat(
                        [table_results[status_name], get_status(table, model_group)], ignore_index=True
                    )
                elif op_metrics_name in csv:
                    table = table.assign(Group=model_group)
                    table_results[op_metrics_name] = pd.concat(
                        [table_results[op_metrics_name], table], ignore_index=True
                    )
            os.chdir(result_file)

        if not table_results[memory_name].empty:
            table_results[memory_over_time_name] = get_memory_over_time(table_results[memory_name])

        if not table_results[latency_name].empty:
            table_results[latency_over_time_name] = get_latency_over_time(args.report_url, table_results[latency_name])

        if not table_results[session_name].empty:
            table_results[session_over_time_name] = get_session_over_time(table_results[session_name])

        if not table_results[status_name].empty:
            table_results[status_over_time_name] = get_status_over_time(table_results[status_name])

        for table in tables:
            print("writing " + table + " to database")
            db_table_name = "ep_model_" + table
            write_table(
                ingest_client,
                args.database,
                table_results[table],
                db_table_name,
                upload_time,
                identifier,
                args.branch,
                args.commit_hash,
                args.commit_datetime,
            )

    except BaseException as e:
        print(str(e))
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
import argparse
import os
from datetime import date

import numpy as np
import onnx
from onnx import AttributeProto, GraphProto, TensorProto, helper, numpy_helper, utils  # noqa: F401


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--output_dir", required=True, help="Path to the build directory.")
    return parser.parse_args()


def write_config(model_dir):
    with open(os.path.join(model_dir, "config.txt"), "w") as f:
        f.write("per_sample_tolerance:1e-3\n")
        f.write("relative_per_sample_tolerance:1e-6\n")


def write_tensor(f, c, input_name=None):
    tensor = numpy_helper.from_array(c)
    if input_name:
        tensor.name = input_name
    body = tensor.SerializeToString()
    f.write(body)


def infer_shapes(model_def):
    onnx.checker.check_model(model_def)
    onnx.helper.strip_doc_string(model_def)
    final_model = onnx.shape_inference.infer_shapes(model_def)
    onnx.checker.check_model(final_model)
    return final_model


def generate_abs_op_test(type, X, top_test_folder):
    for is_raw in [True, False]:
        if is_raw:
            test_folder = os.path.join(top_test_folder, "raw")
        else:
            test_folder = os.path.join(top_test_folder, "not_raw")
        data_dir = os.path.join(test_folder, "test_data_0")
        os.makedirs(data_dir, exist_ok=True)
        # Create one output (ValueInfoProto)
        Y = helper.make_tensor_value_info("Y", type, X.shape)  # noqa: N806
        X_INFO = helper.make_tensor_value_info("X", type, X.shape)  # noqa: N806
        if is_raw:
            tensor_x = onnx.helper.make_tensor(name="X", data_type=type, dims=X.shape, vals=X.tobytes(), raw=True)
        else:
            tensor_x = onnx.helper.make_tensor(name="X", data_type=type, dims=X.shape, vals=X.ravel(), raw=False)
        # Create a node (NodeProto)
        node_def = helper.make_node("Abs", inputs=["X"], outputs=["Y"])

        # Create the graph (GraphProto)
        graph_def = helper.make_graph([node_def], "test-model", [X_INFO], [Y], [tensor_x])
        # Create the model (ModelProto)
        model_def = helper.make_model(graph_def, producer_name="onnx-example")
        # final_model = infer_shapes(model_def)
        final_model = model_def
        if is_raw:
            onnx.external_data_helper.convert_model_to_external_data(final_model, True)
        onnx.save(final_model, os.path.join(test_folder, "model.onnx"))
        expected_output_array = np.abs(X)
        expected_output_tensor = numpy_helper.from_array(expected_output_array)
        with open(os.path.join(data_dir, "output_0.pb"), "wb") as f:
            f.write(expected_output_tensor.SerializeToString())


def generate_size_op_test(type, X, test_folder):
    data_dir = os.path.join(test_folder, "test_data_0")
    os.makedirs(data_dir, exist_ok=True)
    # Create one output (ValueInfoProto)
    Y = helper.make_tensor_value_info("Y", TensorProto.INT64, [])  # noqa: N806
    X_INFO = helper.make_tensor_value_info("X", type, X.shape)  # noqa: N806
    tensor_x = onnx.helper.make_tensor(name="X", data_type=type, dims=X.shape, vals=X.ravel(), raw=False)
    # Create a node (NodeProto)
    node_def = helper.make_node("Size", inputs=["X"], outputs=["Y"])

    # Create the graph (GraphProto)
    graph_def = helper.make_graph([node_def], "test-model", [X_INFO], [Y], [tensor_x])
    # Create the model (ModelProto)
    model_def = helper.make_model(graph_def, producer_name="onnx-example", opset_imports=[helper.make_opsetid("", 18)])
    final_model = infer_shapes(model_def)
    onnx.save(final_model, os.path.join(test_folder, "model.onnx"))
    expected_output_array = np.int64(X.size)
    expected_output_tensor = numpy_helper.from_array(expected_output_array)
    with open(os.path.join(data_dir, "output_0.pb"), "wb") as f:
        f.write(expected_output_tensor.SerializeToString())


def generate_reducesum_op_test(X, test_folder):
    type = onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[X.dtype]
    data_dir = os.path.join(test_folder, "test_data_0")
    os.makedirs(data_dir, exist_ok=True)
    # Create one output (ValueInfoProto)
    Y = helper.make_tensor_value_info("Y", type, [])  # noqa: N806
    X_INFO = helper.make_tensor_value_info("X", type, X.shape)  # noqa: N806
    tensor_x = onnx.helper.make_tensor(name="X", data_type=type, dims=X.shape, vals=X.ravel(), raw=False)
    # Create a node (NodeProto)
    node_def = helper.make_node("ReduceSum", inputs=["X"], outputs=["Y"], keepdims=0)

    # Create the graph (GraphProto)
    graph_def = helper.make_graph([node_def], "test-model", [X_INFO], [Y], [tensor_x])
    # Create the model (ModelProto)
    model_def = helper.make_model(graph_def, producer_name="onnx-example")
    final_model = infer_shapes(model_def)
    onnx.save(final_model, os.path.join(test_folder, "model.onnx"))
    expected_output_array = np.sum(X)
    expected_output_tensor = numpy_helper.from_array(expected_output_array)
    with open(os.path.join(data_dir, "output_0.pb"), "wb") as f:
        f.write(expected_output_tensor.SerializeToString())


def test_abs(output_dir):
    generate_abs_op_test(
        TensorProto.FLOAT,
        np.random.randn(3, 4, 5).astype(np.float32),
        os.path.join(output_dir, "test_abs_float"),
    )
    generate_abs_op_test(
        TensorProto.DOUBLE,
        np.random.randn(3, 4, 5).astype(np.float64),
        os.path.join(output_dir, "test_abs_double"),
    )
    generate_abs_op_test(
        TensorProto.INT8,
        np.int8([-127, -4, 0, 3, 127]),
        os.path.join(output_dir, "test_abs_int8"),
    )
    generate_abs_op_test(
        TensorProto.UINT8,
        np.uint8([0, 1, 20, 255]),
        os.path.join(output_dir, "test_abs_uint8"),
    )
    generate_abs_op_test(
        TensorProto.INT16,
        np.int16([-32767, -4, 0, 3, 32767]),
        os.path.join(output_dir, "test_abs_int16"),
    )
    generate_abs_op_test(
        TensorProto.UINT16,
        np.uint16([0, 3, 32767, 65535]),
        os.path.join(output_dir, "test_abs_uint16"),
    )
    generate_abs_op_test(
        TensorProto.INT32,
        np.int32([-2147483647, -4, 0, 3, 2147483647]),
        os.path.join(output_dir, "test_abs_int32"),
    )
    generate_abs_op_test(
        TensorProto.UINT32,
        np.uint32([0, 1, 20, 4294967295]),
        os.path.join(output_dir, "test_abs_uint32"),
    )
    number_info = np.iinfo(np.int64)
    generate_abs_op_test(
        TensorProto.INT64,
        np.int64([-number_info.max, -4, 0, 3, number_info.max]),
        os.path.join(output_dir, "test_abs_int64"),
    )
    number_info = np.iinfo(np.uint64)
    generate_abs_op_test(
        TensorProto.UINT64,
        np.uint64([0, 1, 20, number_info.max]),
        os.path.join(output_dir, "test_abs_uint64"),
    )


def test_reducesum(output_dir):
    generate_reducesum_op_test(
        np.random.randn(3, 4, 5).astype(np.float32),
        os.path.join(output_dir, "test_reducesum_random"),
    )


def test_size(output_dir):
    generate_size_op_test(
        TensorProto.FLOAT,
        np.random.randn(100, 3000, 10).astype(np.float32),
        os.path.join(output_dir, "test_size_float"),
    )
    generate_size_op_test(
        TensorProto.STRING,
        np.array(["abc", "xy"], dtype=np.bytes_),
        os.path.join(output_dir, "test_size_string"),
    )


args = parse_arguments()
os.makedirs(args.output_dir, exist_ok=True)
# make test values deterministic but variable
today = date.today()
np.random.seed(today.year + today.month + today.day)
test_abs(args.output_dir)
test_size(args.output_dir)
test_reducesum(args.output_dir)
